import re
import six

def get_type_and_data(h_node):
    """ Helper function to return the py_type and data block for a HDF node """
    py_type = h_node.attrs["type"][0]
    data = h_node[()]
#    if h_node.shape == ():
#        data = h_node.value
#    else:
#        data  = h_node[:]
    return py_type, data

def get_type(h_node):
    """ Helper function to return the py_type for a HDF node """
    py_type = h_node.attrs["type"][0]
    return py_type

def sort_keys(key_list):
    """ Take a list of strings and sort it by integer value within string

    Args:
        key_list (list): List of keys

    Returns:
        key_list_sorted (list): List of keys, sorted by integer
    """

    # Py3 h5py returns an irritating KeysView object
    # Py3 also complains about bytes and strings, convert all keys to bytes
    if six.PY3:
        key_list2 = []
        for key in key_list:
            if isinstance(key, str):
                key = bytes(key, 'ascii')
            key_list2.append(key)
        key_list = key_list2

    # Check which keys contain a number
    numbered_keys = [re.search(br'\d+', key) for key in key_list]

    # Sort the keys on number if they have it, or normally if not
    if(len(key_list) and not numbered_keys.count(None)):
        to_int = lambda x: int(re.search(br'\d+', x).group(0))
        return(sorted(key_list, key=to_int))
    else:
        return(sorted(key_list))


def check_is_iterable(py_obj):
    """ Check whether a python object is iterable.

    Note: this treats unicode and string as NON ITERABLE

    Args:
        py_obj: python object to test

    Returns:
        iter_ok (bool): True if item is iterable, False is item is not
    """
    if six.PY2:
        string_types = (str, unicode)
    else:
        string_types = (str, bytes, bytearray)
    if isinstance(py_obj, string_types):
        return False
    try:
        iter(py_obj)
        return True
    except TypeError:
        return False


def check_is_hashable(py_obj):
    """ Check if a python object is hashable

    Note: this function is currently not used, but is useful for future
          development.

    Args:
        py_obj: python object to test
    """

    try:
        py_obj.__hash__()
        return True
    except TypeError:
        return False


def check_iterable_item_type(iter_obj):
    """ Check if all items within an iterable are the same type.

    Args:
        iter_obj: iterable object

    Returns:
        iter_type: type of item contained within the iterable. If
                   the iterable has many types, a boolean False is returned instead.

    References:
    http://stackoverflow.com/questions/13252333/python-check-if-all-elements-of-a-list-are-the-same-type
    """
    iseq = iter(iter_obj)

    try:
        first_type = type(next(iseq))
    except StopIteration:
        return False
    except Exception as ex:
        return False
    else:
        return first_type if all((type(x) is first_type) for x in iseq) else False



# encoding: utf-8
"""
#lookup.py

This file manages all the mappings between hickle/HDF5 metadata and python
types.
There are three dictionaries that are populated by the LoaderManager associated
with each file created or loaded:

1) types_dict
Mapping between python types and dataset and group creation functions, e.g.
    types_dict = {
        list: (create_listlike_dataset, 'list'),
        int: (create_python_dtype_dataset, 'int'),
        np.ndarray: (create_np_array_dataset, 'ndarray'),
        }

2) hkl_types_dict
Mapping between hickle metadata and dataset loading functions, e.g.
    hkl_types_dict = {
        'list': load_list_dataset,
        'tuple': load_tuple_dataset
        }

3) hkl_container_dict
Mapping between hickle metadata and group container classes, e.g.
    hkl_contianer_dict = {
        'list': ListLikeContainer,
        'tuple': TupleLikeContainer,
        'dict': DictLikeContainer
    }

## Extending hickle to add support for other classes and types

The process to add new load/dump capabilities is as follows:

1) Create a file called load_[newmodule/newpackage].py
2) In the load_[newmodule/newpackage].py file, define your create_dataset,
   load_dataset functions and PyContainer objects, along with the 
   'class_register' and 'exclude_register' tables. See the other loaders
   in <Python path>/dist_packages/hickle/loaders directory for examples.
   The columns in register_class table correspond to argument list of 
   hickle.lookup.LoaderManager.register_class method. 
   

3) store the load_[newmodule/newpackage].py file in one of the following locations
   <Python path>/dist_packages/hickle/loaders/
       the loaders supported by hickle globally

   <Python path>/dist_packages/hickle_loaders/
       loaders installed during installation of additional single file python modules
       through pip or legacy os installer

   <Python path>/dist_packages/[newpackage]/hickle_loaders/
       loaders provided by [newpackage] installed through pip or legacy os installer

   [MyPackage]/hickle_loaders/
       loaders specific for objects and classes defined by the modules of [MyPackage] package

   basedir([MyModule|MyApplication])/hickle_loaders
       loaders specific for object and classes defined by [MyModule|MyApplication].py file
       not part of a python package

Loader for a single object or class can also be created by calling 
hickle.lookup.register_class prior to calling hickle.dump and hickle.load

1) create a `create_[MyClass]_dataset` function, a `load_[MyClass]_fcn` function and/or
   `[MyClass]Container` class for [MyClass] class to be dumped and loaded by hickle.
   Examples can be found in the loader modules in <Python path>/dist_packages/hickle/loaders
   directory
2) call `hickle.lookup.LoaderManager.register_class` method as follows from your code

   ```
   from hickle.lookup import LoaderManager

   LoaderManager.register_class(
       [MyClass],'<myclass_type>',
       create_[MyClass]_dataset,
       load_[MyClass]_fcnt, # or None if [MyClass] is mapped to h5py.Group only
       [MyClass]Container , # or None if [MyClass] is mapped to h5py.Dataset only
       True, # if False [MyClass] object will be stored explicitly on any occurrence
       'custom', # set to None to enforce unconditional use of loader
   )

   hickle.dump('my_[MyClass]_object,'myfile.hkl','w',options={'custom':True})
   new_[MyClass]_object = hickle.load('myfile.hkl')
   ```

"""


# %% IMPORTS
# Built-in imports
import sys
import warnings
import types
import re
import weakref
import os.path
from importlib.util import find_spec, module_from_spec,spec_from_file_location,spec_from_loader
from importlib import invalidate_caches

# Package imports
import collections
import pickle
import numpy as np
import h5py as h5

# hickle imports
from .helpers import PyContainer,not_dumpable,nobody_is_my_name,no_compression,NotHicklable
from .loaders import optional_loaders, attribute_prefix

if sys.version_info[:2] <= (3,5): # pragma: no cover
    # define ModuleNotFoundError on __builtins__ to ensure below code is working
    setattr(sys.modules['builtins'],'ModuleNotFoundError',getattr(sys.modules['builtins'],'ModuleNotFoundError',ImportError))

# %% GLOBALS

# %% FUNCTION DEFINITIONS


def load_nothing(h_node, base_type , py_obj_type): # pragma: no cover
    """
    loads nothing
    """
    return nobody_is_my_name

def dump_nothing(py_obj, h_group, name, **kwargs): # pragma: no cover
    """
    dumps nothing
    """
    return nobody_is_my_name


# %% CLASS DEFINITIONS

class _DictItem(): # pragma: no cover
    """
    dummy py_obj for dict_item loader
    """

class NodeReference(): # pragma: no cover
    """
    dummy py_obj_type returned by ReferenceManager.get_type
    when encountering dataset of h5py.ref_dtype  which expose 
    no explicit 'type' attribute.
    """

class ReferenceError(Exception): # pragma: no cover
    """
    exception thrown by ReferenceManager
    """

class LookupError(Exception): # pragma: no cover
    """
    exception thrown if type lookup fails
    """

class SerializedWarning(UserWarning): # pragma: no cover
    """ An object type was not understood

    The data will be serialized using pickle.
    """

class PackageImportDropped(UserWarning): # pragma: no cover
    """
    Package or module defining type/class was removed from 
    sys.modules
    """

class MockedLambdaWarning(UserWarning): # pragma: no cover
    """
    In Python >= 3.8 lambda function fails pickle.loads

    faking restoring lambda to keep legacy hickle 4.x files
    loading properly
    """

class DataRecoveredWarning(UserWarning): # pragma: no cover
    """
    Raised when hickle does not find an appropriate loader for
    a specific type and thus has to fall back to '!recover!' loader
    recover_custom_dataset function or RecoverGroupContainer PyContainer
    object
    """

class AttemptRecoverCustom():
    """
    Dummy type indicating that the data of specific py_obj_type listed in the
    hickle_types_table could not be restored. Most likely pickle.loads encountered
    an ImportError/ModuleNotFoundError or AttributeError indicating that the package
    and/or module defining the py_obj_type is not installed or does not anymore 
    provide the definition of py_object_type.

    In this case hickle tries to at least recover the data and corresponding meta
    data stored in h5py.Group or h5py.Dataset attributes and the base_type string
    indicating the loader used to dump the data.  Only in case this attempt fails
    an exception is thrown.
    """

class RecoveredGroup(dict,AttemptRecoverCustom):
    """
    dict type object representing the content and hickle meta data of a h5py.group
    """

    __slots__ = ('attrs',)

    def __init__(self,*args,attrs={},**kwargs):
        super().__init__(*args,**kwargs)
        self.attrs={name:value for name,value in attrs.items() if name not in {'type'}}

class RecoveredDataset(np.ndarray,AttemptRecoverCustom):
    """
    numpy.ndarray type object representing the content and hickle meta data  of a
    h5py.dataset
    """

    __slots__ = ('attrs',)

    def __new__(cls,input_array,dtype=None,attrs={}):
        array_copy = np.array(input_array,dtype=dtype)
        obj = super().__new__(
            cls,
            shape = array_copy.shape,
            dtype = array_copy.dtype,
            buffer=array_copy,
            offset=0,
            strides=array_copy.strides,
            order = 'C' if array_copy.flags.c_contiguous else 'F'
        )
        obj.attrs = {name:value for name,value in attrs.items() if name not in {'type'}}
        return obj
    
    def __array_finalize__(self,obj):
        if obj is not None:
            self.attrs = getattr(obj,'attrs',{})

class ManagerMeta(type):
    """
    Metaclass for all manager classes derived from the BaseManager class.
    Ensures that the __managers__ class attribute of each immediate child
    class of BaseManager is initialized to a dictionary and ensures that
    it can not be overwritten by any grandchild class and down. The
    __managers__ attribute declared by any grandchild shadowing the
    __managers__ attribute of any of its ancestors is dropped without any
    further notice.
    """

    def __new__(cls, name, bases, namespace, **kwords):
        for _ in ( True for base in bases if isinstance(getattr(base,'__managers__',None),dict) ):
            namespace.pop('__managers__',None)
            break
        else:
            if not isinstance(namespace.get('__managers__',None),dict):
                namespace['__managers__'] = dict() if bases and not object in bases else None
        return super().__new__(cls,name,bases,namespace,**kwords)
                
class BaseManager(metaclass = ManagerMeta):
    """
    Base class providing basic management of simultaneously open managers.
    Must be subclassed.
    """
    __slots__ = ('__weakref__',)

    __managers__ = None

    @classmethod
    def _drop_manager(cls, fileid):
        """
        finalizer callback to properly remove a <manager> object from
        the <manager>.__managers__ structure when corresponding file is
        closed or <manager> object is garbage collected

        Parameters
        ----------
        cls (BaseManager):
            the child <manager> class for which to drop the hdf5 file instance
            referenced to by the specified file id

        fileid (h5py.FileId):
            Id identifying the hdf5 file the ReferenceManager was created for

        """
        try:
            manager = cls.__managers__.get(fileid,None)
            if manager is None:
                return
            cls.__managers__.pop(fileid,None)
        except: # pragma: no cover
            # only triggered in case of race
            pass

    @classmethod
    def create_manager(cls, h_node, create_entry):
        """
        Check whether an instance already exists for the file the h_node belongs to and
        call create_entry if no entry yet exists.

        Parameters
        ----------
        cls:
            the manager class to create a new instance for

        h_node (h5py.File, h5py.Group, h5py.Dataset):
            the h5py node or its h_root_group or file to create a new ReferenceManager
            object for.

        create_entry (callable):
            function or method which returns a new table entry. A table entry
            is a tuple or list with contains as its first item the newly
            created <manager> object. It may include further items specific to
            the actual BaseManager subclass.
                
        Raises
        ------
        LookupError:
            if manager has already been created for h_node or its h_root_group 
                
        """
        manager = cls.__managers__.get(h_node.file.id,None)
        if manager is not None:
            raise LookupError(
                "'{}' type manager already created for file '{}'".format(
                cls.__name__,h_node.file.filename
            )
        )
        table = cls.__managers__[h_node.file.id] = create_entry()
        weakref.finalize(table[0],cls._drop_manager,h_node.file.id)
        return table[0]

    @classmethod
    def get_manager(cls, h_node):
        """
        return manager responsible for the file containing h_node 
    
        Parameters
        ----------
        h_node (h5py.File, h5py.Group, h5py.Dataset):
            the h5py node to obtain the responsible manager for.
            
        Raises
        ------
        LookupError:
            if no manager has been created yet for h_node or its h_root_group 
        """
        try:
            return cls.__managers__[h_node.file.id][0]
        except KeyError:
            raise ReferenceError("no managers exist for file '{}'".format(h_node.file.filename))

    def __init__(self):
        if type.mro(self.__class__)[0] is BaseManager:
            raise TypeError("'BaseManager' class must be subclassed")

    def __enter__(self):
        raise NotImplementedError("'{}' type object must implement Python ContextManager protocol")

    def __exit__(self, exc_type, exc_value, exc_traceback, h_node=None):
        # remove this ReferenceManager object from the table of active ReferenceManager objects
        # and cleanly unlink from any h5py object instance and id references managed. Finalize
        # 'hickle_types_table' overlay if it was created by __init__ for hickle 4.0.X file
        self.__class__._drop_manager(h_node.file.id)

class ReferenceManager(BaseManager, dict):
    """
    Manages all object and type references created for basic
    and type special memoisation.

    To create a ReferenceManager call ReferenceManager.create_manager
    function. The value returned can be and shall be used within a
    with statement to ensure it is garbage collected before file is closed.
    For example:

        with ReferenceManager.create_manager(h_root_group) as memo:
            _dump(data,h_root_group,'data',memo,loader,**kwargs)

        with ReferenceManager.create_manager(h_root_group) as memo:
            _load(py_container,'data',h_root_group['data'],memo,loader)

        with ReferenceManager.create_manager(h_root_group,fix_lambda_obj_type) as memo:
            _load(py_container,'data',h_root_group['data'],memo,loader)

    NOTE: for creating appropriate loader object see LoaderManager
    """

    __slots__ = (
        '_py_obj_type_table', # hickle_types_table h5py.Group storing type information
        '_py_obj_type_link', # dictionary linking py_obj_type and representation in hickle_types_table
        '_base_type_link', # dictionary linking base_type string and representation in hickle_types_table
        '_overlay', # in memory hdf5 dummy file hosting dummy hickle_types_table for hickle 4.x files
        'pickle_loads' # reference to pickle.loads method
    )

    
    @staticmethod
    def get_root(h_node):
        """
        returns the h_root_group the passed h_node belongs to.
        """

        # try to resolve the 'type' attribute of the h_node
        entry_ref = h_node.attrs.get('type',None)
        if isinstance(entry_ref,h5.Reference):

            # return the grandparent of the referenced py_obj_type dataset as it
            # also the h_root_group of h_node
            try:
                entry = h_node.file.get(entry_ref,None)
            except ValueError: # pragma: no cover
                entry = None
            if entry is not None:
                return entry.parent.parent
        if h_node.parent == h_node.file:

            # h_node is either the h_root_group it self or the file node representing
            # the open hickle file. 
            return h_node if isinstance(h_node,h5.Group) else h_node.file

        # either h_node has not yet a 'type' assigned or contains pickle string
        # which has implicit b'pickle' type. try to resolve h_root_group from its
        # parent 'type' entry if any
        entry_ref = h_node.parent.attrs.get('type',None)
        if not isinstance(entry_ref,h5.Reference):
            if entry_ref is None:

                # parent has neither a 'type' assigned
                return h_node if isinstance(h_node,h5.Group) else h_node.file

            # 'type' seems to be a byte string or string fallback to h_node.file
            return h_node.file
        try:
            entry = h_node.file.get(entry_ref,None)
        except ValueError: # pragma: no cover
            entry = None
        if entry is None:

            # 'type' reference seems to be stale
            return h_node if isinstance(h_node,h5.Group) else h_node.file

        # return the grand parent of the referenced py_obj_type dataset as it
        # is also the h_root_group of h_node
        return  entry.parent.parent

    @staticmethod
    def _drop_overlay(h5file):
        """
        closes in memory overlay file providing dummy 'hickle_types_table' structure
        for hdf5 files which were created by hickle 4.x 
        """
        h5file.close()

    @classmethod
    def create_manager(cls,h_node, pickle_loads = pickle.loads):
        """
        creates a new ReferenceManager object for the h_root_group the h_node
        belongs to. 


        Parameters
        ----------
        h_node (h5py.Group, h5py.Dataset):
            the h5py node or its h_root_group to create a new ReferenceManager
            object for.

        pickle_loads (FunctionType,MethodType):
            method to be used to expand py_obj_type pickle strings.
            defaults to pickle.loads. Must be set to fix_lambda_obj_type 
            for hickle file created by hickle 4.x.

        Raises
        ------
        LookupError:
            if ReferenceManager has already been created for h_node or its h_root_group 
        """

        def create_manager():
            return ( 
                ReferenceManager(h_node,pickle_loads = pickle_loads),
                ReferenceManager.get_root(h_node)
            )
        return super().create_manager(h_node,create_manager)

    def __init__(self, h_root_group, *args,pickle_loads = pickle.loads, **kwargs):
        """
        constructs ReferenceManager object

        Parameters
        ----------
        h_root_group (h5py.Group):
            see ReferenceManager.create_manager

        args (tuple,list):
            passed to dict.__init__

        pickle_loads (FunctionType,MethodType):
            see ReferenceManager.create_manager

        kwargs (dict):
            passed to dict.__init__

        Raises
        ------
        ReferenceError:
            In case an error occurs while loading 'hickle_types_table' from an
            existing file opened for reading and writing
        """
        super().__init__(*args,**kwargs)
        self._py_obj_type_link = dict()
        self._base_type_link = dict()
        self._overlay = None
        self.pickle_loads = pickle_loads

        # get the 'hickle_types_table' member of h_root_group or create it anew
        # in case none found. In case hdf5 file is opened for reading only
        # create an in memory hdf5 file (managed by hdf5 'core' driver) providing
        # an empty dummy hickle_types_table. This is necessary to ensue that
        # ReferenceManager.resolve_type works properly on hickle 4.x files which
        # store type information directly in h5py.Group and h5py.Datasets attrs
        # structure.
        self._py_obj_type_table = h_root_group.get('hickle_types_table',None)
        if self._py_obj_type_table is None:
            if h_root_group.file.mode == 'r+':
                self._py_obj_type_table = h_root_group.create_group("hickle_types_table",track_order = True)
            else:
                h5_overlay = h5.File(
                    '{}.hover'.format(h_root_group.file.filename.rsplit('.',1)[0]),
                    mode='w', driver='core',backing_store=False
                )
                self._py_obj_type_table = h5_overlay.create_group("hickle_types_table",track_order = True)
                self._overlay = weakref.finalize(self,ReferenceManager._drop_overlay,h5_overlay)
            return

        # verify that '_py_obj_type_table' is a valid h5py.Group object
        if not isinstance(self._py_obj_type_table,h5.Group):
            raise ReferenceError("'hickle_types_table' invalid: Must be HDF5 Group entry")

        # if h_root_group.file was opened for writing restore '_py_obj_type_link' and
        # '_base_type_link' table entries from '_py_obj_type_table' to ensure when
        # h5py.Group and h5py.Dataset are added anew to h_root_group tree structure
        # their 'type' attribute is set to the correct py_obj_type reference by the
        # ReferenceManager.store_type method. Each of '_py_obj_type_link' and
        # '_base_type_link' tables can be used to properly restore  the 'py_obj_type'
        # and 'base_type' when loading the file as well as assigning to the 'type'
        # attribute the appropriate 'py_obj_type' dataset reference from the
        # '_py_obj_type_table' when dumping data to the file.
        if h_root_group.file.mode != 'r+':
            return
        for _, entry in self._py_obj_type_table.items():
            if entry.shape is None and entry.dtype == 'S1':
                base_type = entry.name.rsplit('/',1)[-1].encode('ascii')
                self._base_type_link[base_type] = entry
                self._base_type_link[entry.id] = base_type
                continue
            base_type_ref = entry.attrs.get('base_type',None)
            if not isinstance(base_type_ref,h5.Reference):
                raise ReferenceError(
                    "inconsistent 'hickle_types_table' entries for py_obj_type '{}': "
                    "no base_type".format(py_obj_type)
                )
            try:
                base_type_entry = entry.file.get(base_type_ref,None)
            except ValueError: # pragma: no cover
                base_type_entry = None
            if base_type_entry is None:
                raise ReferenceError(
                    "inconsistent 'hickle_types_table' entries for py_obj_type '{}': "
                    "stale base_type".format(py_obj_type)
                )
            base_type = self._base_type_link.get(base_type_entry.id,None)
            if base_type is None:
                base_type = base_type_entry.name.rsplit('/',1)[-1].encode('ascii')
            try:
                py_obj_type = pickle.loads(entry[()])
            except (ImportError,AttributeError):
                py_obj_type = AttemptRecoverCustom
                entry_link = py_obj_type,'!recover!',base_type
            else:
                entry_link = py_obj_type,base_type
                self._py_obj_type_link[id(py_obj_type)] = entry
            self._py_obj_type_link[entry.id] = entry_link

    def store_type(self, h_node, py_obj_type, base_type = None, attr_name = 'type', **kwargs):
        """
        assigns a 'py_obj_type' entry reference to the attribute specified by attr_name
        of h_node and creates if not present the appropriate 'hickle_types_table'
        entries for py_obj_type and base_type.

        Note
        ----
            Storing and restoring the content of nodes containing pickle byte strings
            is fully managed by pickle.dumps and pickle.loads functions including
            selection of appropriate py_obj_type. Therefore no explicit entry for 
            object and b'pickle' py_obj_type and base_type pairs indicating pickled
            content of pickled dataset are created.

        Parameters
        ----------
        h_node (h5py.Group, h5py.Dataset):
            node the 'type' attribute a 'hickle_types_table' entry corresponding to 
            the provided py_obj_type, base_type entry pair shall be assigned to.

        py_obj_type (any type or class):
            the type or class of the object instance represented by h_node

        base_type (bytes):
            the base-type bytes string of the loader used to create the h_node and
            restore an object instance form on load. If None no 'hickle_types_table'
            will be created for py_obj_type if not already present and a LookupError
            exception is raised instead.

        attr_name (str):
            the name of the attribute the type reference shall be stored to. Defaults
            to 'type'

        kwargs (dict):
            keyword arguments to be passed to h5py.Group.create_dataset function
            when creating the entries for py_obj_type and base_type anew

        Raises
        ------
        ValueError:
            if base_type is not a valid bytes string

        LookupError:
            if base_type is None and no 'hickle_types_table' entry exists for
            py_obj_type yet

        """

        # return immediately if py_obj_type is object as h_node contains pickled byte
        # string of the actual object dumped
        if py_obj_type is object:
            return

        # if no entry within the 'hickle_types_table' exists yet
        # for py_obj_type create the corresponding pickle string dataset
        # and store appropriate entries in the '_py_obj_type_link' table for
        # further use by ReferenceManager.store_type and ReferenceManager.resolve_type
        # methods
        py_obj_type_id = id(py_obj_type)
        entry = self._py_obj_type_link.get(py_obj_type_id,None)
        if entry is None:
            if base_type is None:
                raise LookupError(
                    "no entry found for py_obj_type '{}'".format(py_obj_type.__name__)
                ) 
            if not isinstance(base_type,(str,bytes)) or not base_type:
                raise ValueError("base_type must be non empty bytes string")
            type_entry = memoryview(pickle.dumps(py_obj_type))
            type_entry = np.array(type_entry,copy = False)
            type_entry.dtype = 'S1'
            entry = self._py_obj_type_table.create_dataset(
                str(len(self._py_obj_type_table)),
                data=type_entry,
                shape=(1,type_entry.size),
                **kwargs
            )
            # assign a reference to base_type entry within 'hickle_types_table' to
            # the 'base_type' attribute of the newly created py_obj_type entry.
            # if 'hickle_types_table' does not yet contain empty dataset entry for
            # base_type create it and store appropriate entries in the '_base_type_link' table
            # for further use by ReferenceManager.store_type and ReferenceManager,resolve_type
            # methods
            base_entry = self._base_type_link.get(base_type,None)
            if base_entry is None:
                base_entry = self._base_type_link[base_type] = self._py_obj_type_table.create_dataset(
                    base_type.decode('ascii'),
                    shape=None,dtype = 'S1',
                    **no_compression(kwargs)
                )
                self._base_type_link[base_entry.id] = base_type
            entry.attrs['base_type'] = base_entry.ref
            self._py_obj_type_link[py_obj_type_id] = entry
            self._py_obj_type_link[entry.id] = (py_obj_type,base_type)
        h_node.attrs[attr_name] = entry.ref

    def resolve_type(self,h_node,attr_name = 'type',base_type_type = 1):
        """
        resolves the py_obj_type and base_type pair referenced to by the 'type' attribute and
        if present the 'base_type' attribute.

        Note: If the 'base_type' attribute is present it is assumed that the dataset was
            created by hickle 4.x version. Consequently it is assumed that the 'type' attribute
            contains a pickle bytes string to load the py_obj_type from instead of a reference
            to a 'hickle_types_table' entry representing the py_obj_type base_type pair of h_node.

        Note: If 'type' attribute is not present h_node represents either a h5py.Reference to
            the actual node of the object to be restored or contains a pickle bytes string.
            In either case the corresponding implicit py_obj_type base_type pair 
            (NodeReference, b'!node-reference!') or (object,b'pickle') respective is assumed and
            returned.

        Note: If restoring 'py_object_type' from pickle string stored in type attribute or 
            'hickle_types_table' fails than implicit py_obj_type base_type pair 
            (AttemptRecoverCustom,'!recover!') is returned instead of the actual 'py_obj_type' 
            base_type pair is returned. The latter can be retrieved by setting 'base_type_type'
            to 2 in this case

        Parameters
        ----------
        h_node (h5py.Group,h5py.Dataset):
            the node to resolve py_obj_type and base_type for using reference stored in
            attribute specified by attr_name

        attr_name (str):
            the name of the attribute the type reference shall be restored from. Defaults
            to and must be 'type' in case not a h5py.Reference.

        base_type_type (int):
            1 (default) base_type used to select loader
           -1 original base_type corresponding to not understood py_obj_type of
              recovered h5py.Group or h5py.Dataset

        Returns
        -------
        tuple containing (py_obj_type,base_type,is_container)

        py_obj_type:
            the python type of the restored object

        base_type:
            the base_type string indicating the loader to be used for properly
            restoring the py_obj_type instance or the base_type string 

        is_container:
            boolean flag indicating whether h_node represents a h5py.Group or
            h5py.Reference both of which have to be handled by corresponding
            PyContainer type loaders or a h5py.Dataset for which the appropriate
            load_fn is to be called.
        """

        # load the type attribute indicated by attr_name. If not present check if h_node
        # is h5py.Reference dataset or a dataset containing a pickle bytes
        # string. In either case assume (NodeReference,b'!node-reference!') or (object,b'pickle')
        # respective as (py_obj_type, base_type) pair and set is_container flag to True for
        # h5py.Reference and False otherwise.
        # 
        # NOTE: hickle 4.x legacy file does not store 'type' attribute for h5py.Group nodes with 
        #       b'dict_item' base_type. As h5py.Groups do not have a dtype attribute the check
        #       whether h_node.dtype equals h5py.ref_dtype will raise AttributeError.
        #       If h_node represents a b'dict_item' than self.pickle_loads will point to
        #       fix_lambda_obj_type below which will properly handle None value of type_ref
        #       in any other case file is not a hickle 4.x legacy file and thus has to be
        #       considered broken
        type_ref = h_node.attrs.get(attr_name,None)
        if not isinstance(type_ref,h5.Reference):
            if type_ref is None:
                try:
                    metadata = h_node.dtype.metadata
                except (AttributeError,):
                    pass
                else:
                    if metadata is not None and issubclass(metadata.get('ref',object),h5.Reference):
                        return NodeReference,b'!node-reference!',True
                    return object,b'pickle',False

            # check if 'type' attribute of h_node contains a reference to a 'hickle_types_table'
            # entry. If not use pickle to restore py_object_type from the 'type' attribute value
            # directly if possible
            try:

                # set is_container_flag to True if h_node is h5py.Group type object and false
                # otherwise
                return self.pickle_loads(type_ref), h_node.attrs.get('base_type', b'pickle'), isinstance(h_node, h5.Group)
            except (ModuleNotFoundError,AttributeError):
                # module missing or py_object_type not provided by module
                return AttemptRecoverCustom,( h_node.attrs.get('base_type',b'pickle') if base_type_type == 2 else b'!recover!' ),isinstance(h_node,h5.Group)
            except (TypeError, pickle.UnpicklingError, EOFError):
                raise ReferenceError(
                    "node '{}': '{}' attribute ('{}')invalid: not a pickle byte string".format(
                        h_node.name,attr_name,type_ref
                    )
                )
        try:
            entry = self._py_obj_type_table[type_ref]
        except (ValueError, KeyError):
            raise ReferenceError(
                "node '{}': '{}' attribute invalid: stale reference".format(
                    h_node.name,attr_name
                )
            )

        # load (py_obj_type,base_type) pair from _py_obj_type_link for 'hickle_types_table' entry
        # referenced by 'type' entry. Create appropriate _py_obj_type_link and _base_type_link
        # entries if if not present for (py_obj_type,base_type) pair for further use by
        # ReferenceManager.store_type and ReferenceManager.resolve_type methods.
        type_info = self._py_obj_type_link.get(entry.id, None)
        if type_info is None:
            base_type_ref = entry.attrs.get('base_type', None)
            if base_type_ref is None:
                base_type = b'pickle'
            else:
                try:
                    base_type_entry = self._py_obj_type_table[base_type_ref]
                except ( ValueError,KeyError ):
                    # TODO should be recovered here instead?
                    raise ReferenceError(
                        "stale base_type reference encountered for '{}' type table entry".format(
                            entry.name
                        )
                    )
                base_type = self._base_type_link.get(base_type_entry.id,None)
                if base_type is None:

                    # get the relative table entry name form full path name of entry node
                    base_type = base_type_entry.name.rsplit('/',1)[-1].encode('ASCII')
                    self._base_type_link[base_type] = base_type_entry
                    self._base_type_link[base_type_entry.id] = base_type
            try:
                py_obj_type = self.pickle_loads(entry[()])
            except (ModuleNotFoundError,AttributeError):
                py_obj_type = AttemptRecoverCustom
                entry_link = (py_obj_type,b'!recover!',base_type)
            else:
                entry_link = (py_obj_type,base_type)
                self._py_obj_type_link[id(py_obj_type)] = entry
            type_info = self._py_obj_type_link[entry.id] = entry_link
        # return (py_obj_type,base_type). set is_container flag to true if 
        # h_node is h5py.Group object and false otherwise
        return (type_info[0],type_info[base_type_type],isinstance(h_node,h5.Group))

    def __enter__(self):
        if not isinstance(self._py_obj_type_table, h5.Group) or not self._py_obj_type_table:
            raise RuntimeError(
                "Stale ReferenceManager, call ReferenceManager.create_manager to create a new one"
            )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not isinstance(self._py_obj_type_table, h5.Group) or not self._py_obj_type_table:
            return

        # remove this ReferenceManager object from the table of active ReferenceManager objects
        # and cleanly unlink from any h5py object instance and id references managed. Finalize
        # 'hickle_types_table' overlay if it was created by __init__ for hickle 4.x file
        super().__exit__(exc_type, exc_value, exc_traceback, self._py_obj_type_table)
        self._py_obj_type_table = None
        self._py_obj_type_link = None
        self._base_type_link = None
        self.pickle_loads = None
        if self._overlay is not None:
            self._overlay()
            self._overlay = None

#####################
# loading optional  #
#####################

_managed_by_hickle = {'hickle', ''}

_custom_loader_enabled_builtins = {'__main__':('','')}

class LoaderManager(BaseManager):
    """
    Handles the file specific lookup of loader to be used to dump or load
    a python object of a specific type 

    To create a LoaderManager call LoaderManager.create_manager
    function. The value returned can be and shall be used within a
    with statement for example as follows:

        with LoaderManager.create_manager(h_root_group) as loader:
            _dump(data,h_root_group,'data',memo,loader,**kwargs)

        with LoaderManager.create_manager(h_root_group,False,{'custom':true}) as loader:
            _load(py_container,'data',h_root_group['data'],memo,loader)

        with LoaderManager.create_manager(h_root_group,True) as memo:
            _load(py_container,'data',h_root_group['data'],memo,loader)

    NOTE: for creating appropriate memo object see ReferenceManager

    """

    # Define dict of all acceptable types dependent upon loader option
    __py_types__ = { 
        None: {},
        'hickle-4.x': {},
        **{ option:{} for option in optional_loaders }
    }

    # Define dict of all acceptable load function dependent upon loader option
    __hkl_functions__ = {
        None: {},
        'hickle-4.x': {},
        **{ option:{} for option in optional_loaders }
    }

    # Define dict of all acceptable hickle container types dependent upon loader option
    __hkl_container__ = {
        None: {},
        'hickle-4.x': {},
        **{ option:{} for option in optional_loaders }
    }

    # Empty list (hashtable) of loaded loader names
    __loaded_loaders__ = set()


    @classmethod
    def register_class(
        cls, myclass_type, hkl_str, dump_function=None, load_function=None,
        container_class=None, memoise = True, option=None
    ):
        """
        Register a new class to be recognized and dumped or restored by hickle.
    
        Parameters
        ----------
        myclass_type (type.class):
            the class to register dump_fcn, load_fcn, PyContainer for

        hkl_str (str):
            String to write to HDF5 file identifying class and loader suitable
            for restoring the py_object described by the data stored in hickle
            file.

            NOTE: dict_item, pickle, !node-refrence!, !recover! and any other
                string enclosed within a pair of !! can not be modified if once
                registered. Strings quoted by !! must be added as global loader
                with option = None.
                
        dump_function (callable):
            callable to write data to HDF5

        load_function (callable):
            function to load data from HDF5

        container_class (PyContainer):
            PyContainer type proxy class to load data from HDF5

        memoise (bool): 
            True: references to the object instances of class shall be remembered
                during dump and load for properly resolving multiple references
                to the same object instance.

            False: every occurrence of an instance of the object has to be dumped
                and restored on load disregarding instances already present.

        option (str, None):
            String identifying set of loaders which shall only be used when
            specific feature or category is requested on top of global loaders.
            If None than loader is globally to be used if there is no other loader
            registered for myclass_type. 

            NOTE: only strings listed in 'optional_loaders' exported by 
                  hickle.loaders.__init__ and 'hickle-4.x' are accepted. 
                
        Raises
        ------
        TypeError:
            loader for myclass_type may only be registered by hickle core modules not
            loaded from hickle/loaders/ directory, <packagepath>/hickle_loaders/,
            <modulepath>/hickle_loaders/ or <__main__path>/hickle_loaders/ directory
            by explicitly calling LoaderManager.register_class method.

        ValueError:
            if optional loader modules tries to shadow 'dict_item', 'pickle' and any
            loader marked as essential to proper function of hickle.dump and hickle.load 
            by ! prefix and postfix ('!node-reference!', '!recover!').

        LookupError:
            if optional loader denoted by option is unknown. Any new option must be
            listed in 'optional_loaders' exported by 'hickle.loaders.__init__.py'
            file to be recognized as valid option
        """
    
        if (
            myclass_type is object or
            isinstance(
                myclass_type,
                (types.FunctionType, types.BuiltinFunctionType, types.MethodType, types.BuiltinMethodType)
            ) or
            issubclass(myclass_type,(type,_DictItem))
        ):
            # object, all functions, methods, class objects and the special _DictItem class
            # type objects are to be handled by hickle core only. 
            dump_module = getattr(dump_function, '__module__', '').split('.', 2)
            load_module = getattr(load_function, '__module__', '').split('.', 2)
            container_module = getattr(container_class, '__module__', '').split('.', 2)
            if {dump_module[0], load_module[0], container_module[0]} - _managed_by_hickle:
                raise TypeError(
                    "loader for '{}' type managed by hickle only".format(
                        myclass_type.__name__
                    )
                )
            if "loaders" in {*dump_module[1:2], *load_module[1:2], *container_module[1:2]}:
                raise TypeError(
                    "loader for '{}' type managed by hickle core only".format(
                        myclass_type.__name__
                    )
                )
        if ( 
            ( cls.__hkl_functions__[None].get(hkl_str) or cls.__hkl_container__[None].get(hkl_str) ) and
            ( hkl_str[:1] == hkl_str[-1:] == b'!' or hkl_str in disallow_in_option )
        ):
            raise ValueError(
                "'{}' base_type may not be shadowed by loader".format(hkl_str)
            )
        # add loader
        try:
            if dump_function is not None:
                cls.__py_types__[option][myclass_type] = ( dump_function, hkl_str,memoise)
            if load_function is not None:
                cls.__hkl_functions__[option][hkl_str] = load_function
                cls.__hkl_functions__[option][hkl_str.decode('ascii')] = load_function
            if container_class is not None:
                cls.__hkl_container__[option][hkl_str] = container_class
                cls.__hkl_container__[option][hkl_str.decode('ascii')] = container_class
        except KeyError:
            raise LookupError("Invalid option '{}' encountered".format(option))
        
            


    @classmethod
    def register_class_exclude(cls, hkl_str_to_ignore, option = None):
        """
        Tell loading function to ignore any HDF5 dataset with attribute 'type=XYZ'
    
        Parameters
        ----------
        hkl_str_to_ignore (str): attribute type=string to ignore and exclude
            from loading.

        option (str, None):
            String identifying set of optional loaders from which class shall
            be excluded
                
        Raises
        ------
        ValueError:
            class is managed by hickle core machinery and thus may not be ignored

        LookupError:
            option loader shall belong to is unknown. Any new option must be listed
            in 'optional_loaders' exported by 'hickle.loaders.__init__.py' file to be
            recognized as valid option
        """
    
        if hkl_str_to_ignore[0] == hkl_str_to_ignore[-1] == b'!' or hkl_str_to_ignore in disallowed_to_ignore:
            raise ValueError(
                "excluding '{}' base_type managed by hickle core not possible".format(
                    hkl_str_to_ignore
                )
            )
        try:
            cls.__hkl_functions__[option][hkl_str_to_ignore] = load_nothing
            cls.__hkl_container__[option][hkl_str_to_ignore] = NoContainer
            cls.__hkl_functions__[option][hkl_str_to_ignore.decode('ascii')] = load_nothing
            cls.__hkl_container__[option][hkl_str_to_ignore.decode('ascii')] = NoContainer
        except KeyError:
            raise LookupError("'{}' option unknown".format(option))

    __slots__ = ( 'types_dict', 'hkl_types_dict', 'hkl_container_dict', '_mro', '_file')


    _option_formatter = '{}{{}}'.format(attribute_prefix)
    _option_parser = re.compile(r'^{}(.*)$'.format(attribute_prefix),re.I)

    def __init__(self, h_root_group, legacy = False, options = None):
        """
        constructs LoaderManager object

        Parameters
        ----------
        h_root_group (h5py.Group):
            see LoaderManager.create_manager

        legacy (bool):
            If true the file h_node belongs to is in legacy hickle 4.x format.
            Ensure lambda py_obj_type strings are loaded properly and
            'hickle-4.x' type loaders are included within types_dict, 
            'hkl_types_dict' and 'hkl_container_dict'

        options (dict):
            optional loaders to be loaded. Each key names one loader and
            its value indicates whether to be used (True) or excluded (False) 

        Raises
        ------
        LookupError:
            option loader unknown
        """

        # initialize lookup dictionaries with set of common loaders
        self.types_dict = collections.ChainMap(self.__class__.__py_types__[None])
        self.hkl_types_dict = collections.ChainMap(self.__class__.__hkl_functions__[None])
        self.hkl_container_dict = collections.ChainMap(self.__class__.__hkl_container__[None])

        # Select source of optional loader flags. If option is None try to read options
        # from h_root_group.attrs structure. Otherwise use content of options dict store
        # each entry to be used within h_root_group.attrs structure or update entry there
        if options is None:
            option_items = ( 
                match.group(1).lower() 
                for match,on in ( 
                    ( LoaderManager._option_parser.match(name), value )
                    for name, value in h_root_group.attrs.items() 
                ) 
                if match and on
            )
        else:
            def set_option_items():
                for option_key,on in options.items():
                    if not on:
                        continue
                    h_root_group.attrs[LoaderManager._option_formatter.format(option_key.upper())] = on 
                    yield option_key
            option_items = set_option_items()
        # try to include loader set indicated by option_name
        try:
            for option_name in option_items:
                self.types_dict.maps.insert(0,self.__class__.__py_types__[option_name])
                self.hkl_types_dict.maps.insert(0,self.__class__.__hkl_functions__[option_name])
                self.hkl_container_dict.maps.insert(0,self.__class__.__hkl_container__[option_name])
        except KeyError:
            raise LookupError("Option '{}' invalid".format(option_name))
            
        # add loaders required to properly load legacy files created by hickle 4.x and
        # ensure that non class types are properly reported by load_loader
        if legacy:
            self._mro = type_legacy_mro
            self.types_dict.maps.insert(0,self.__class__.__py_types__['hickle-4.x'])
            self.hkl_types_dict.maps.insert(0,self.__class__.__hkl_functions__['hickle-4.x'])
            self.hkl_container_dict.maps.insert(0,self.__class__.__hkl_container__['hickle-4.x'])
        else:
            self._mro = type.mro
        self._file = h_root_group.file
        
    def load_loader(self, py_obj_type,*,base_type=None):
        """
        Checks if given `py_obj` requires an additional loader to be handled
        properly and loads it if so.
    
        Parameters
        ----------
        py_obj:
            the Python object to find an appropriate loader for
    
        Returns
        -------
        tuple containing (py_obj, (create_dataset, base_type, memoise))

        py_obj:
             the Python object the loader was requested for
    
        (create_dataset,base_type,memoise):
             tuple providing create_dataset function, name of base_type used
             to represent py_obj and the boolean memoise flag indicating
             whether loaded object shall be remembered for restoring further
             references to it or must be loaded every time encountered.
    
        Raises
        ------
        RuntimeError:
            in case py object is defined by hickle core machinery.
        """
    
        types_dict = self.types_dict
        loaded_loaders = self.__class__.__loaded_loaders__
        # loop over the entire mro_list of py_obj_type
        for mro_item in self._mro(py_obj_type):

            # Check if mro_item is already listed in types_dict and return if found 
            loader_item = types_dict.get(mro_item,None)
            if loader_item is not None:
                return py_obj_type,loader_item
    
            # Obtain the package name of mro_item
            package_list = mro_item.__module__.split('.',2)
    
            package_file = None 
            if package_list[0] == 'hickle':
                if package_list[1] != 'loaders':
                    if base_type is not None and ( base_type in self.hkl_types_dict or base_type in self.hkl_container_dict):
                        return py_obj_type,(not_dumpable,base_type,True)
                    print(mro_item,package_list)
                    raise RuntimeError(
                        "objects defined by hickle core must be registered"
                        " before first dump or load"
                    )
                if (
                    len(package_list) < 3 or
                    not package_list[2].startswith("load_") or
                    '.' in package_list[2][5:]
                ):
                    warnings.warn(
                        "ignoring '{!r}' dummy type not defined by loader module".format(py_obj_type),
                        RuntimeWarning
                    )
                    continue

                # dummy objects are not dumpable ensure that future lookups return that result
                loader_item = types_dict.get(mro_item,None)
                if loader_item is None:
                    loader_item = types_dict[mro_item] = ( not_dumpable, b'NotHicklable',False )

                # ensure module of mro_item is loaded as loader as it will contain
                # loader which knows how to handle group or dataset with dummy as 
                # py_obj_type 
                loader_name = mro_item.__module__
                if loader_name in loaded_loaders:

                    # loader already loaded as triggered by dummy abort search and return
                    # what found so far as fallback to further bases does not make sense
                    return py_obj_type,loader_item
            else:
                loader_name,package_file = _custom_loader_enabled_builtins.get(package_list[0],(None,''))
                if loader_name is None:
                    # construct the name of the associated loader
                    loader_name = 'hickle.loaders.load_{:s}'.format(package_list[0])
                elif not loader_name:
                    # try to resolve module name for __main__ script and other generic modules
                    package_module = sys.modules.get(package_list[0],None)
                    if package_module is None:
                        warnings.warn(
                            "package/module '{}' defining '{}' type dropped".format(
                                package_list[0],mro_item.__name__
                            ),
                            PackageImportDropped
                        )
                        continue
                    package_file = getattr(package_module,'__file__',None)
                    if package_file is None:
                        package_loader = getattr(package_module,'__loader__',None)
                        if package_loader is None: # pragma: no cover
                            # just to secure against "very smart" tinkering
                            # with python import machinery, no serious testable use-case known and expected
                            continue
                        package_spec = spec_from_loader(package_list[0],package_loader)
                        if not getattr(package_spec,'has_location',False):
                            continue
                        package_file = package_spec.origin
                    if not os.path.isabs(package_file): # pragma: no cover
                        # not sure if this case wouldn't just be result of "very smart" tinkering
                        # with python import machinery, no serious testable use-case known yet
                        package_spec = find_spec(os.path.basename(package_file.rsplit('.')[0]))
                        if not getattr(package_spec,'has_location',False): # pargma: no cover
                            # not sure if this case wouldn't just be result of "very smart" tinkering
                            # with python import machinery, no serious testable use-case known yet
                            continue
                        package_file = package_spec.origin
                    package_list[0],allow_custom_loader = os.path.basename(package_file).rsplit('.')[0],package_list[0]
                    loader_name = 'hickle.loaders.load_{:s}'.format(package_list[0])
                    _custom_loader_enabled_builtins[allow_custom_loader] = loader_name, package_file
    
                # Check if this module is already loaded
                if loader_name in loaded_loaders:
                    # loader is loaded but does not define loader for mro_item
                    # check next base class
                    continue
    
            # check if loader module has already been loaded. If use that instead
            # of importing it anew
            loader = sys.modules.get(loader_name,None)
            if loader is None:
                # Try to load a loader with this name
                loader_spec = find_spec(loader_name)
                if loader_spec is None:
                    assert isinstance(package_file,str), "package_file name for _custom_loader_enabled_builtins must be string"
                    if not package_file:
                        package_spec = getattr(sys.modules.get(package_list[0],None),'__spec__',None)
                        if package_spec is None:
                            package_spec = find_spec(package_list[0])
                        if not getattr(package_spec,'has_location',False):
                            # can't resolve package or base module hosting mro_item
                            continue
                        package_file = package_spec.origin
                    package_path = os.path.dirname(package_file)
                    package_loader_path = os.path.join(
                        package_path, "hickle_loaders", "load_{:s}.py".format(package_list[0])
                    )
                    try:
                        fid = open(package_loader_path,'rb')
                    except FileNotFoundError:
                        try:
                            package_loader_path += 'c'
                            fid = open(package_loader_path,'rb')
                        except FileNotFoundError:
                            # no file for loader module found
                            continue
                        else:
                            fid.close()
                    else:
                        fid.close()
                    loader_spec = spec_from_file_location(loader_name,package_loader_path)
                # import the the loader module described by module_spec
                # any import errors and exceptions result at this stage from
                # errors inside module and not cause loader module does not
                # exist
                loader = module_from_spec(loader_spec)
                loader_spec.loader.exec_module(loader)
                sys.modules[loader_name] = loader
    
            # load all loaders defined by loader module
            for next_loader in loader.class_register:
                self.register_class(*next_loader)
            for drop_loader in ( loader if isinstance(loader,(list,tuple)) else (loader,None) for loader in loader.exclude_register) :
                self.register_class_exclude(*drop_loader)
            loaded_loaders.add(loader_name)
    
            # check if loader module defines a loader for base_class mro_item
            loader_item = types_dict.get(mro_item,None)
            if loader_item is None:
                # the new loader does not define loader for mro_item
                # check next base class
                continue

            # return loader for base_class mro_item
            return py_obj_type,loader_item
    
        # no appropriate loader found. Lower py_object_type to object and
        # return fallback to pickle
        return object,(create_pickled_dataset,b'pickle',True)

    @classmethod
    def create_manager(cls, h_node, legacy = False, options = None):
        """
        creates an new LoaderManager object for the h_root_group the h_node
        belongs to. 


        Parameters
        ----------
        h_node (h5py.Group, h5py.Dataset):
            the h5py node or its h_root_group to create a new LoaderManager
            object for.

        legacy (bool):
            if true file h_node belongs to is in legacy hickle 4.x format
            ensure lambda py_obj_type strings are loaded properly and
            'hickle-4.x' type loaders are included within types_dict, 
            'hkl_types_dict' and 'hkl_container_dict'

        options (dict):
            optional loaders to be loaded. Each key names one loader and
            its value indicates whether to be used (True) or excluded (False) 

        Raises
        ------
        LookupError:
            if ReferenceManager has already been created for h_node or its h_root_group 
        """

        def create_manager():
            return (LoaderManager(h_node,legacy,options),)
        return super().create_manager(h_node,create_manager)

    def __enter__(self):
        if not isinstance(self._file,h5.File) or not self._file:
            raise RuntimeError(
                "Stale LoaderManager, call LoaderManager.create_manager to create a new one"
            )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not isinstance(self._file,h5.File) or not self._file:
            return
        super().__exit__(exc_type, exc_value, exc_traceback, self._file)
        self._file = None
        self._mro = None
        self.types_dict = None
        self.hkl_types_dict = None
        self.hkl_container_dict = None
        
def type_legacy_mro(cls):
    """
    drop in replacement of type.mro for loading legacy hickle 4.x files which were
    created without generalized PyContainer objects available. Consequently some
    h5py.Datasets and h5py.Group objects expose function objects as their py_obj_type
    type.mro expects classes only.

    Parameters
    ----------
    cls (type):
        the py_obj_type/class of the object to load or dump

    Returns
    -------
        mro list for cls as returned by type.mro  or in case cls is a function or method
        a single element tuple is returned
    """
    if isinstance(
        cls,
        (types.FunctionType,types.BuiltinFunctionType,types.MethodType,types.BuiltinMethodType)
    ):
        return (cls,)
    return type.mro(cls) 

# %% BUILTIN LOADERS (not maskable)

# list of below hkl_types which may not be ignored
# NOTE: types which are enclosed in !! pair are disallowed in any case
disallowed_to_ignore = {b'dict_item', b'pickle' }

# list of below hkl_types which may not be redefined by optional loader
# NOTE: types which are enclosed in !! pair are disallowed in any case
disallow_in_option = {b'pickle'}

class NoContainer(PyContainer): # pragma: no cover
    """
    load nothing container
    """

    def convert(self):
        pass


class _DictItemContainer(PyContainer):
    """
    PyContainer reducing hickle version 4.x dict_item type h5py.Group to 
    its content for inclusion within dict h5py.Group
    """

    def convert(self):
        return self._content[0]

LoaderManager.register_class(
    _DictItem, b'dict_item', dump_nothing, load_nothing, _DictItemContainer, False, 'hickle-4.x'
)

        
class ExpandReferenceContainer(PyContainer):
    """
    PyContainer for properly restoring additional references to an object
    instance shared multiple times within the dumped object structure
    """

    def filter(self,h_parent):
        """
        resolves the h5py.Reference link and yields the the node
        it refers to as sub item of h_parent so that it can be
        properly loaded by recursively calling hickle._load method
        independent whether it can be directly loaded from the memo
        dictionary or has to be restored from file. 
        """

        try:
            referred_node = h_parent.file.get(h_parent[()],None)
        except ( ValueError, KeyError ): # pragma no cover
            referred_node = None
        if referred_node is None:
            raise ReferenceError("node '{}' stale node reference".format(h_parent.name))
        yield referred_node.name.rsplit('/',1)[-1], referred_node

    def convert(self):
        """
        returns the object the reference was pointing to
        """

        return self._content[0]

# objects created by resolving h5py.Reference datasets are already stored inside
# memo dictionary so no need to memoise them.
LoaderManager.register_class(
    NodeReference, b'!node-reference!', dump_nothing, load_nothing, ExpandReferenceContainer, False
)


def create_pickled_dataset(py_obj, h_group, name, reason = None, **kwargs):
    """
    Create pickle string as object can not be mapped to any other h5py
    structure.

    Parameters
    ----------
    py_obj:
        python object to dump; default if item is not matched.

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    reason (str,None):
        reason why py_object has to be pickled eg. string
        provided by NotHicklable exception

    Warnings
    -------
    SerializedWarning:
        issued before pickle string is created

    """

    # for what ever reason py_obj could not be successfully reduced
    # ask pickle for help and report to user.
    reason_str = " (Reason: %s)" % (reason) if reason is not None else ""
    warnings.warn(
        "{!r} type not understood, data is serialized:{:s}".format(
            py_obj.__class__.__name__, reason_str
        ),
        SerializedWarning
    )

    # store object as pickle string
    pickled_obj = pickle.dumps(py_obj)
    d = h_group.create_dataset(name, data = memoryview(pickled_obj), **kwargs)
    return d,() 

def load_pickled_data(h_node, base_type, py_obj_type):
    """
    loade pickle string and return resulting py_obj
    """
    try:
        return pickle.loads(h_node[()])
    except (ImportError,AttributeError):
        return RecoveredDataset(h_node[()],dtype = h_node.dtype,attrs = dict(h_node.attrs))

        
# no dump method is registered for object as this is the default for
# any unknown object and for classes, functions and methods
LoaderManager.register_class(object,b'pickle',None,load_pickled_data)

def recover_custom_dataset(h_node,base_type,py_obj_type):
    """
    drop in load_fcn for any base_type no appropriate loader could be found
    """
    manager = ReferenceManager.get_manager(h_node)
    _,base_type,_ = manager.resolve_type(h_node,base_type_type = -1)
    warnings.warn(
        "loader '{}' missing for '{}' type object. Data recovered ({})".format(
            base_type, 
            py_obj_type.__name__ if not isinstance(py_obj_type, AttemptRecoverCustom) else None,
            h_node.name.rsplit('/')[-1]
        ),
        DataRecoveredWarning
    )
    attrs = dict(h_node.attrs)
    attrs['base_type'] = base_type
    return RecoveredDataset(h_node[()],dtype=h_node.dtype,attrs=attrs)

class RecoverGroupContainer(PyContainer):
    """
    drop in PyContainer for any base_type not appropriate loader could be found
    """
    def __init__(self,h5_attrs, base_type, object_type):
        super().__init__(h5_attrs, base_type, object_type,_content = {})
    
    def filter(self,h_parent):
        """
        switch base_type to the one loader is missing for
        """

        warnings.warn(
            "loader '{}' missing for '{}' type object. Data recovered ({})".format(
                self.base_type,
                self.object_type.__name__ if not isinstance(self.object_type,AttemptRecoverCustom) else None,
                h_parent.name.rsplit('/')[-1]
            ),
            DataRecoveredWarning
        )
        manager = ReferenceManager.get_manager(h_parent)
        _,self.base_type,_ = manager.resolve_type(h_parent,base_type_type = -1)
        yield from h_parent.items()

    def append(self,name,item,h5_attrs):
        if isinstance(item,AttemptRecoverCustom):
            self._content[name] = item
        else:
            self._content[name] = (item,{ key:value for key,value in h5_attrs.items() if key not in {'type'}})

    def convert(self):
        attrs = {key:value for key,value in self._h5_attrs.items() if key not in {'type'}}
        attrs['base_type'] = self.base_type
        return RecoveredGroup(self._content,attrs=attrs)


LoaderManager.register_class(AttemptRecoverCustom,b'!recover!',None,recover_custom_dataset,RecoverGroupContainer,True)
    

def _moc_numpy_array_object_lambda(x):
    """
    drop in replacement for lambda object types which seem not
    any more be accepted by pickle for Python 3.8 and onward.
    see fix_lambda_obj_type function below

    Parameters
    ----------
    x (list):
        itemlist from which to return first element

    Returns
    -------
    first element of provided list
    """
    return x[0]

LoaderManager.register_class(
    _moc_numpy_array_object_lambda, b'!moc_lambda!', dump_nothing, load_nothing, None, True, 'hickle-4.x'
)

def fix_lambda_obj_type(bytes_object, *, fix_imports=True, encoding="ASCII", errors="strict"):
    """
    drop in replacement for pickle.loads method when loading files created by hickle 4.x 
    It captures any TypeError thrown by pickle.loads when encountering a pickle string 
    representing a lambda function used as py_obj_type for a h5py.Dataset or h5py.Group.
    While in Python <3.8 pickle loads creates the lambda Python >= 3.8 throws an 
    error when encountering such a pickle string. This is captured and
    _moc_numpy_array_object_lambda returned instead. Further some h5py.Group and h5py.Datasets
    do not provide any py_obj_type for them object is returned assuming that proper loader has
    been identified by other objects already
    """
    if bytes_object is None:
        return object
    try:
        return pickle.loads(bytes_object, fix_imports=fix_imports, encoding=encoding, errors=errors)
    except TypeError:
        warnings.warn(
            "presenting '{!r}' instead of stored lambda 'type'".format(
                _moc_numpy_array_object_lambda
            ),
            MockedLambdaWarning
        )
        return _moc_numpy_array_object_lambda



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle_lookup.py

Unit tests for hickle module -- lookup functions.

"""

# %% IMPORTS
import pytest
import sys
import shutil
import types
import weakref
import compileall
import os

# Package imports
import re
import collections
import numpy as np
import h5py
import pickle
from importlib.util import find_spec,spec_from_loader,spec_from_file_location
from importlib import reload
from copy import copy
import os.path
from py.path import local

# hickle imports
from hickle.helpers import PyContainer,not_dumpable
from hickle.loaders import optional_loaders, attribute_prefix
import hickle.lookup as lookup
    
# Set current working directory to the temporary directory
local.get_temproot().chdir()

# %% DATA DEFINITIONS

dummy_data = (1,2,3)


# %% FIXTURES

@pytest.fixture
def h5_data(request):
    """
    create dummy hdf5 test data file for testing PyContainer, H5NodeFilterProxy and
    ReferenceManager. Uses name of executed test as part of filename
    """

    dummy_file = h5py.File('hickle_lookup_{}.hdf5'.format(request.function.__name__),'w')
    filename = dummy_file.filename
    test_data = dummy_file.create_group("root_group")
    yield test_data
    if dummy_file:
        dummy_file.close()

@pytest.fixture()
def loader_table():

    """
    create a class_register and a exclude_register table for testing
    register_class and register_class_exclude functions

    0: dataset only loader
    1: PyContainer only loader
    2: not dumped loader
    3: external loader module trying to overwrite hickle core loader
    4: hickle loader module trying to overload hickle core loader
    3: loader defined by hickle core
    
    """

    # clear loaded_loaders, types_dict, hkl_types_dict and hkl_contianer_dict
    # to ensure no loader preset by hickle core or hickle loader module
    # intervenes with test
    global lookup
    lookup.LoaderManager.__loaded_loaders__.clear()
    tuple( True for opt in lookup.LoaderManager.__py_types__.values() if opt.clear() )
    tuple( True for opt in lookup.LoaderManager.__hkl_functions__.values() if opt.clear() )
    tuple( True for opt in lookup.LoaderManager.__hkl_container__.values() if opt.clear() )

    # simulate loader definitions found within loader modules
    def create_test_dataset(myclass_type,h_group,name,**kwargs):
        return h_group,()

    def load_test_dataset(h_node,base_type,py_obj_type):
        return 12

    class TestContainer(PyContainer):
        def convert(self):
            return self._content[0]

    class NotHicklePackage(TestContainer):
        """
        checks if container_class provided by module outside 
        hickle package  tries to define alternative loader for
        IteratorProxy class handled by hickle core directly
        """
        __module__ = "nothickle.loaders.load_builtins"

    class HickleLoadersModule(TestContainer):
        """
        checks if container_class provided by 
        hickle.loaders module tries to define alternative loader for
        IteratorProxy class handled by hickle core directly
        """
        __module__ = "hickle.loaders.load_builtins"

    class IsHickleCore(TestContainer):
        """
        Simulates loader registered by hickle.hickle module
        """
        __module__ = "hickle.hickle"

    # provide the table
    yield [
        (int,b'int',create_test_dataset,load_test_dataset,None,False),
        (list,b'list',create_test_dataset,None,TestContainer,True),
        (tuple,b'tuple',None,load_test_dataset,TestContainer),
        (lookup._DictItem,b'dict_item',None,None,NotHicklePackage),
        (lookup._DictItem,b'pickle',None,None,HickleLoadersModule),
        (lookup._DictItem,b'dict_item',lookup.LoaderManager.register_class,None,IsHickleCore)
    ]

    # cleanup and reload hickle.lookup module to reset it to its initial state
    # in case hickle.hickle has already been preloaded by pytest also reload it
    # to ensure no side effects occur during later tests
    lookup.LoaderManager.__loaded_loaders__.clear()
    tuple( True for opt in lookup.LoaderManager.__py_types__.values() if opt.clear() )
    tuple( True for opt in lookup.LoaderManager.__hkl_functions__.values() if opt.clear() )
    tuple( True for opt in lookup.LoaderManager.__hkl_container__.values() if opt.clear() )
    
    reload(lookup)
    lookup = sys.modules[lookup.__name__]
    hickle_hickle = sys.modules.get("hickle.hickle",None)
    if hickle_hickle is not None:
        reload(hickle_hickle)

# %% CLASS DEFINITIONS

class ToBeInLoadersOrNotToBe():
    """
    Dummy class used to check that only loaders for Python objects
    are accepted by load_loader which are either declared
    outside hickle or are pre registered by hickle core through directly
    calling register_class or are declared by a load_<module>.py module
    within the pickle.loaders package

    Also it is used in simulating reduced object tuple with all trailing
    None items removed
    """
    __slots__ = ()

    def __reduce_ex__(self,proto = pickle.DEFAULT_PROTOCOL):
        reduced = super(ToBeInLoadersOrNotToBe,self).__reduce_ex__(proto)
        for index,item in enumerate(reduced[:1:-1],0):
            if item is not None:
                return reduced[:(-index if index > 0 else None)]
        return reduced

    def __reduce__(self):
        reduced = super(ToBeInLoadersOrNotToBe,self).__reduce__()
        for index,item in enumerate(reduced[:1:-1],0):
            if item is not None:
                return reduced[:(-index if index > 0 else None)]
        return reduced

    def __eq__(self,other):
        return other.__class__ is self.__class__

    def __ne__(self,other):
        return self != other
class ClassToDump():
    """
    Primary class used to test create_pickled_dataset function
    """
    def __init__(self,hallo,welt,with_default=1):
        self._data = hallo,welt,with_default

    def __eq__(self,other):
        return other.__class__ is self.__class__ and self._data == other._data

    def __ne__(self,other):
        return self != other

class ClassToDumpCompact(ClassToDump):
    """
    Class which may be handled by 'compact_expand' loader
    """
    def __compact__(self):
        return self._data

    def __expand__(self,compact):
        self._data = compact

class ClassToDumpCompactOff(ClassToDump):
    """
    Class which enforces that any instance is pickled
    independent whether 'compact_expand' loader was selected
    for hickle.dump call or not
    """
    def __compact__(self):
        return None

class ClassToDumpCompactStrange(ClassToDump):
    """
    Class which does not properly implement
    '__compact__' and '__expand__' methods
    recommended by compact expand protocol
    """
    def __compact__(self):
        return self._data

class ClassToDumpCompactStrange2(ClassToDump):
    """
    Another class which does not properly implement
    '__compact__' and '__expand__' methods
    recommended by compact expand protocol
    """
    def __compact__(self):
        return 42
    
class ClassToDumpCompactDataset(ClassToDump):
    """
    Class which is to be represented by a h5py.Dataset
    instead of a h5py.Group in its compacted form
    """
    def __compact__(self):
        return "{}|{}|{}".format(*self._data)

    def __expand__(self,compact):
        self._data = compact.split("|")
        self._data[2] = int(self._data[2])
        self._data = (*self._data,)

class SimpleClass():
    """
    simple class used to check that instance __dict__ is properly dumped and
    restored by create_pickled_dataset and PickledContainer
    """
    def __init__(self):
        self.someattr = "I'm some attr"
        self.someother = 12

    def __eq__(self,other):
        return other.__class__ is self.__class__ and self.__dict__ == other.__dict__

    def __ne__(self,other):
        return self != other

class NoExtendList(list):
    """
    special list class used to test whether append is properly used
    when list like object is dumped and restored through create_pickled_dataset
    and PickledContainer
    """

    def __getattribute__(self,name):
        if name == "extend":
            raise AttributeError("no extend")
        return super(NoExtendList,self).__getattribute__(name)

# %% FUNCTION DEFINITIONS

def function_to_dump(hallo,welt,with_default=1):
    """
    non class function to be dumped and restored through
    create_pickled_dataset and load_pickled_data
    """
    return hallo,welt,with_default

def load_anyhing(h_node, base_type , py_obj_type): # pragma: no cover
    """
    loads nothing
    """
    return None

def test_AttemptRecoverCustom_classes(h5_data):
    recovered_group = lookup.RecoveredGroup({'hello':1},attrs={'world':2,'type':42})
    assert recovered_group == {'hello':1} and recovered_group.attrs == {'world':2}
    array_to_recover = np.random.random_sample([4,2])
    dataset_to_recover = h5_data.create_dataset('to_recover',data=array_to_recover)
    dataset_to_recover.attrs['world'] = 2
    dataset_to_recover.attrs['type'] = 42
    recovered_dataset = lookup.RecoveredDataset(dataset_to_recover[()],dtype=dataset_to_recover.dtype,attrs=dataset_to_recover.attrs)
    assert np.allclose(recovered_dataset,array_to_recover)
    assert recovered_dataset.dtype == array_to_recover.dtype
    assert recovered_dataset.attrs == {'world':2}
    #recovered = lookup.recover_custom_dataset(dataset_to_recover,'unknown',dataset_to_recover.attrs['type']) 
    #assert recovered.dtype == array_to_recover.dtype and recovered == array_to_recover
    #assert recovered.attrs == {'world':2}
    

def test_LoaderManager_register_class(loader_table):
    """
    tests the register_class method
    """

    # try to register dataset only loader specified by loader_table
    # and retrieve its contents from types_dict and hkl_types_dict
    loader_spec = loader_table[0]
    lookup.LoaderManager.register_class(*loader_spec)
    assert lookup.LoaderManager.__py_types__[None][loader_spec[0]] == (*loader_spec[2:0:-1],loader_spec[5])
    assert lookup.LoaderManager.__hkl_functions__[None][loader_spec[1]] == loader_spec[3]
    with pytest.raises(KeyError):
        lookup.LoaderManager.__hkl_container__[None][loader_spec[1]] is None

    # try to register PyContainer only loader specified by loader_table
    # and retrieve its contents from types_dict and hkl_container_dict
    loader_spec = loader_table[1]
    lookup.LoaderManager.register_class(*loader_spec)
    assert lookup.LoaderManager.__py_types__[None][loader_spec[0]] == (*loader_spec[2:0:-1],loader_spec[5])
    with pytest.raises(KeyError):
        lookup.LoaderManager.__hkl_functions__[None][loader_spec[1]] is None
    assert lookup.LoaderManager.__hkl_container__[None][loader_spec[1]] == loader_spec[4]


    # try to register container without dump_function specified by
    # loader table and try to retrieve load_function and PyContainer from
    # hkl_types_dict and hkl_container_dict
    loader_spec = loader_table[2]
    lookup.LoaderManager.register_class(*loader_spec)
    with pytest.raises(KeyError):
        lookup.LoaderManager.__py_types__[None][loader_spec[0]][1] == loader_spec[1]
    assert lookup.LoaderManager.__hkl_functions__[None][loader_spec[1]] == loader_spec[3]
    assert lookup.LoaderManager.__hkl_container__[None][loader_spec[1]] == loader_spec[4]

    # try to register loader shadowing loader preset by hickle core
    # defined by external loader module
    loader_spec = loader_table[3]
    with pytest.raises(TypeError,match = r"loader\s+for\s+'\w+'\s+type\s+managed\s+by\s+hickle\s+only"):
        lookup.LoaderManager.register_class(*loader_spec)
    loader_spec = loader_table[4]

    # try to register loader shadowing loader preset by hickle core
    # defined by hickle loaders module
    with pytest.raises(TypeError,match = r"loader\s+for\s+'\w+'\s+type\s+managed\s+by\s+hickle\s+core\s+only"):
        lookup.LoaderManager.register_class(*loader_spec)

    # simulate registering loader preset by hickle core
    loader_spec = loader_table[5]
    lookup.LoaderManager.register_class(*loader_spec)
    loader_spec = loader_table[0]
    lookup.LoaderManager.__hkl_functions__[None][b'!node-reference!'] = loader_spec[3:5]
    with pytest.raises(ValueError):
        lookup.LoaderManager.register_class(loader_spec[0],b'!node-reference!',*loader_spec[2:],'custom')
    lookup.LoaderManager.__hkl_functions__[None].pop(b'!node-reference!')
    with pytest.raises(lookup.LookupError):
        lookup.LoaderManager.register_class(*loader_spec,'mine')

def test_LoaderManager_register_class_exclude(loader_table):
    """
    test register class exclude function
    """

    # try to disable loading of loader preset by hickle core
    base_type = loader_table[5][1]
    lookup.LoaderManager.register_class(*loader_table[2])
    lookup.LoaderManager.register_class(*loader_table[5])
    with pytest.raises(ValueError,match = r"excluding\s+'.+'\s+base_type\s+managed\s+by\s+hickle\s+core\s+not\s+possible"):
        lookup.LoaderManager.register_class_exclude(base_type)

    # disable any of the other loaders
    base_type = loader_table[2][1]
    lookup.LoaderManager.register_class_exclude(base_type)
    with pytest.raises(lookup.LookupError):
        lookup.LoaderManager.register_class_exclude(base_type,'compact')


def patch_importlib_util_find_spec(name,package=None):
    """
    function used to temporarily redirect search for loaders
    to hickle_loader directory in test directory for testing
    loading of new loaders
    """
    return find_spec("hickle.tests." + name.replace('.','_',1),package)

def patch_importlib_util_find_spec_no_load_builtins(name,package=None):
    """
    function used to temporarily redirect search for loaders
    to hickle_loader directory in test directory for testing
    loading of new loaders
    """
    if name in {'hickle.loaders.load_builtins'}:
        return None
    return find_spec("hickle.tests." + name.replace('.','_',1),package)

def patch_importlib_util_spec_from_tests_loader(name, loader, *, origin=None, is_package=None):
    """
    function used to temporarily redirect search for loaders
    to hickle_loader directory in test directory for testing
    loading of new loaders
    """
    name = name.replace('.','_',1)
    myloader = copy(sys.modules['hickle.tests'].__loader__)
    myloader.name = "hickle.tests." + name
    myloader.path = os.path.join(os.path.dirname(myloader.path),'{}.py'.format(name))
    return spec_from_loader(myloader.name,myloader,origin=origin,is_package=is_package)

def patch_importlib_util_spec_from_loader(name, loader, *, origin=None, is_package=None):
    """
    function used to temporarily redirect search for loaders
    to hickle_loader directory in test directory for testing
    loading of new loaders
    """
    return spec_from_loader("hickle.tests." + name.replace('.','_',1),loader,origin=origin,is_package=is_package)

def patch_importlib_util_spec_from_file_location(name, location, *, loader=None, submodule_search_locations=None):
    """
    function used to temporarily redirect search for loaders
    to hickle_loader directory in test directory for testing
    loading of new loaders
    """
    return spec_from_file_location("hickle.tests." + name.replace('.','_',1),location,loader=loader,submodule_search_locations =submodule_search_locations)


def patch_importlib_util_find_no_spec(name,package=None):
    """
    function used to simulate situation where no appropriate loader 
    could be found for object
    """
    return None

def patch_importlib_util_no_spec_from_loader(name, loader, *, origin=None, is_package=None):
    """
    function used to simulate situation where no appropriate loader 
    could be found for object
    """
    return None

def patch_importlib_util_no_spec_from_file_location(name, location, *, loader=None, submodule_search_locations=None):
    """
    function used to simulate situation where no appropriate loader 
    could be found for object
    """
    return None
def patch_hide_collections_loader(name,package=None):
    if name in ('hickle.loaders.load_collections'):
        return None
    return find_spec(name,package)

def test_LoaderManager(loader_table,h5_data):
    """
    tests LoaderManager constructor
    """
    manager = lookup.LoaderManager(h5_data,False)
    assert isinstance(manager.types_dict,collections.ChainMap)
    assert manager.types_dict.maps[0] is lookup.LoaderManager.__py_types__[None]
    assert isinstance(manager.hkl_types_dict,collections.ChainMap)
    assert manager.hkl_types_dict.maps[0] is lookup.LoaderManager.__hkl_functions__[None]
    assert isinstance(manager.hkl_container_dict,collections.ChainMap)
    assert manager.hkl_container_dict.maps[0] is lookup.LoaderManager.__hkl_container__[None]
    assert manager._mro is type.mro
    assert manager._file.id == h5_data.file.id
    manager = lookup.LoaderManager(h5_data,True)
    assert manager.types_dict.maps[0] is lookup.LoaderManager.__py_types__['hickle-4.x']
    assert manager.types_dict.maps[1] is lookup.LoaderManager.__py_types__[None]
    assert manager.hkl_types_dict.maps[0] is lookup.LoaderManager.__hkl_functions__['hickle-4.x']
    assert manager.hkl_types_dict.maps[1] is lookup.LoaderManager.__hkl_functions__[None]
    assert manager.hkl_container_dict.maps[0] is lookup.LoaderManager.__hkl_container__['hickle-4.x']
    assert manager.hkl_container_dict.maps[1] is lookup.LoaderManager.__hkl_container__[None]
    assert manager._mro is lookup.type_legacy_mro
    assert manager._file.id == h5_data.file.id

    ###### amend #####
    manager = lookup.LoaderManager(h5_data,False,{'custom':True})
    assert manager.types_dict.maps[0] is lookup.LoaderManager.__py_types__['custom']
    assert manager.types_dict.maps[1] is lookup.LoaderManager.__py_types__[None]
    assert manager.hkl_types_dict.maps[0] is lookup.LoaderManager.__hkl_functions__['custom']
    assert manager.hkl_types_dict.maps[1] is lookup.LoaderManager.__hkl_functions__[None]
    assert manager.hkl_container_dict.maps[0] is lookup.LoaderManager.__hkl_container__['custom']
    assert manager.hkl_container_dict.maps[1] is lookup.LoaderManager.__hkl_container__[None]
    assert manager._file.id == h5_data.file.id
    assert h5_data.attrs.get('{}CUSTOM'.format(attribute_prefix),None)
    manager = lookup.LoaderManager(h5_data,False,None)
    assert manager.types_dict.maps[0] is lookup.LoaderManager.__py_types__['custom']
    assert manager.types_dict.maps[1] is lookup.LoaderManager.__py_types__[None]
    assert manager.hkl_types_dict.maps[0] is lookup.LoaderManager.__hkl_functions__['custom']
    assert manager.hkl_types_dict.maps[1] is lookup.LoaderManager.__hkl_functions__[None]
    assert manager.hkl_container_dict.maps[0] is lookup.LoaderManager.__hkl_container__['custom']
    assert manager.hkl_container_dict.maps[1] is lookup.LoaderManager.__hkl_container__[None]
    assert manager._file.id == h5_data.file.id
    h5_data.attrs.pop('{}CUSTOM'.format(attribute_prefix),None)
    manager = lookup.LoaderManager(h5_data,False,{'custom':False})
    assert manager.types_dict.maps[0] is lookup.LoaderManager.__py_types__[None]
    assert manager.hkl_types_dict.maps[0] is lookup.LoaderManager.__hkl_functions__[None]
    assert manager.hkl_container_dict.maps[0] is lookup.LoaderManager.__hkl_container__[None]
    assert h5_data.attrs.get('{}CUSTOM'.format(attribute_prefix),h5_data) is h5_data

    with pytest.raises(lookup.LookupError):
        manager = lookup.LoaderManager(h5_data,False,{'compact':True})


def test_LoaderManager_drop_manager(h5_data):
    """
    test static LoaderManager._drop_table method
    """
    loader = lookup.LoaderManager(h5_data)
    lookup.LoaderManager.__managers__[h5_data.file.id] = (loader,)
    some_other_file = h5py.File('someother.hdf5','w')
    some_other_root = some_other_file.create_group('root')
    lookup.LoaderManager._drop_manager(some_other_root.file.id)
    lookup.LoaderManager.__managers__[some_other_file.file.id] = (lookup.LoaderManager(some_other_root),)
    assert lookup.LoaderManager.__managers__.get(h5_data.file.id,None) == (loader,)
    lookup.LoaderManager._drop_manager(h5_data.file.id)
    assert lookup.LoaderManager.__managers__.get(h5_data.file.id,None) is None
    lookup.LoaderManager._drop_manager(some_other_root.file.id)
    assert not lookup.LoaderManager.__managers__
    some_other_file.close()
    
def test_LoaderManager_create_manager(h5_data):
    """
    test public static LoaderManager.create_manager function
    """
    second_tree = h5_data.file.create_group('seondary_root')
    loader = lookup.LoaderManager.create_manager(h5_data)
    assert lookup.LoaderManager.__managers__[h5_data.file.id][0] is loader
    with pytest.raises(lookup.LookupError):
        second_table = lookup.LoaderManager.create_manager(second_tree)
    lookup.LoaderManager._drop_manager(h5_data.file.id)

def test_LoaderManager_context(h5_data):
    """
    test use of LoaderManager as context manager
    """
    with lookup.LoaderManager.create_manager(h5_data) as loader:
        assert lookup.LoaderManager.__managers__[h5_data.file.id][0] is loader
    assert loader._file is None
    with pytest.raises(RuntimeError):
        with loader as loader2:
            pass
    loader.__exit__(None,None,None)

def test_LoaderManager_load_loader(loader_table,h5_data,monkeypatch):
    """
    test LoaderManager.load_loader method
    """

    # some data to check loader for
    # assume loader should be load_builtins loader
    py_object = dict()
    loader_name = "hickle.loaders.load_builtins"
    with monkeypatch.context() as moc_import_lib:
        with lookup.LoaderManager.create_manager(h5_data) as loader:

            # hide loader from hickle.lookup.loaded_loaders and check that 
            # fallback loader for python object is returned
            moc_import_lib.setattr("importlib.util.find_spec",patch_importlib_util_find_no_spec)
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_importlib_util_find_no_spec)
            moc_import_lib.setattr("importlib.util.spec_from_loader",patch_importlib_util_no_spec_from_loader)
            moc_import_lib.setattr("hickle.lookup.spec_from_loader",patch_importlib_util_no_spec_from_loader)
            moc_import_lib.setattr("importlib.util.spec_from_file_location",patch_importlib_util_no_spec_from_file_location)
            moc_import_lib.setattr("hickle.lookup.spec_from_file_location",patch_importlib_util_no_spec_from_file_location)
            moc_import_lib.delitem(sys.modules,"hickle.loaders.load_builtins",raising=False)
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is object and nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)

            lookup._custom_loader_enabled_builtins[py_obj_type.__class__.__module__] = ('','')
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is object and nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
            
            backup_builtins = sys.modules['builtins']
            moc_import_lib.delitem(sys.modules,'builtins')
            with pytest.warns(lookup.PackageImportDropped):# TODO when warning is added run check for warning
                py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is object and nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
            moc_import_lib.setitem(sys.modules,'builtins',backup_builtins)


            # redirect load_builtins loader to tests/hickle_loader path
            moc_import_lib.setattr("importlib.util.spec_from_file_location",patch_importlib_util_spec_from_file_location)
            moc_import_lib.setattr("hickle.lookup.spec_from_file_location",patch_importlib_util_spec_from_file_location)
            #py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            #assert py_obj_type is dict and nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
            moc_import_lib.setattr("importlib.util.find_spec",patch_importlib_util_find_spec)
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_importlib_util_find_spec)
            moc_import_lib.setattr("importlib.util.spec_from_loader",patch_importlib_util_spec_from_loader)
            moc_import_lib.setattr("hickle.lookup.spec_from_loader",patch_importlib_util_spec_from_loader)
    
            # try to find appropriate loader for dict object, a mock of this
            # loader should be provided by hickle/tests/hickle_loaders/load_builtins
            # module ensure that this module is the one found by load_loader function
            import hickle.tests.hickle_loaders.load_builtins as load_builtins
            moc_import_lib.setitem(sys.modules,loader_name,load_builtins)
            moc_import_lib.setattr("importlib.util.spec_from_loader",patch_importlib_util_spec_from_tests_loader)
            moc_import_lib.setattr("hickle.lookup.spec_from_loader",patch_importlib_util_spec_from_tests_loader)
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is dict and nopickleloader == (load_builtins.create_package_test,b'dict',True)

            # simulate loading of package or local loader from hickle_loaders directory
            backup_load_builtins = sys.modules.pop('hickle.loaders.load_builtins',None)
            backup_py_obj_type = loader.types_dict.pop(dict,None)
            backup_loaded_loaders = lookup.LoaderManager.__loaded_loaders__.discard('hickle.loaders.load_builtins')
            moc_import_lib.setattr("importlib.util.find_spec",patch_importlib_util_find_spec_no_load_builtins)
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_importlib_util_find_spec_no_load_builtins)
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is dict 
            assert nopickleloader == (sys.modules['hickle.loaders.load_builtins'].create_package_test,b'dict',True)
            ## back to start test successful fallback to legacy .pyc in case no source is available for package
            sys.modules.pop('hickle.loaders.load_builtins','None')
            loader.types_dict.pop(dict,None)
            lookup.LoaderManager.__loaded_loaders__.discard('hickle.loaders.load_builtins')
            pyc_path = load_builtins.__file__ + 'c'
            if not os.path.isfile(pyc_path):
                compileall.compile_file(load_builtins.__file__,legacy=True)
                assert os.path.isfile(pyc_path)
            base_dir,base_name = os.path.split(load_builtins.__file__)
            hidden_source = os.path.join(base_dir,'.{}h'.format(base_name))
            os.rename(load_builtins.__file__,hidden_source)
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is dict 
            assert nopickleloader == (sys.modules['hickle.loaders.load_builtins'].create_package_test,b'dict',True)
            #once again just checking that if no legacy .pyc next base is tried
            sys.modules.pop('hickle.loaders.load_builtins','None')
            loader.types_dict.pop(dict,None)
            lookup.LoaderManager.__loaded_loaders__.discard('hickle.loaders.load_builtins')
            os.remove(pyc_path)
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is object
            assert nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
            os.rename(hidden_source,load_builtins.__file__)
            moc_import_lib.setattr("importlib.util.spec_from_loader",patch_importlib_util_spec_from_loader)
            moc_import_lib.setattr("hickle.lookup.spec_from_loader",patch_importlib_util_spec_from_loader)
            moc_import_lib.setattr("importlib.util.find_spec",patch_importlib_util_find_spec)
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_importlib_util_find_spec)
            sys.modules['hickle.loaders.load_builtins'] = backup_load_builtins
            loader.types_dict[dict] = backup_py_obj_type
            # not added by missing legacy .pyc test re-add manually here
            lookup.LoaderManager.__loaded_loaders__.add('hickle.loaders.load_builtins')
            lookup._custom_loader_enabled_builtins.pop(py_obj_type.__class__.__module__,None)
    
            # preload dataset only loader and check that it can be resolved directly
            loader_spec = loader_table[0]
            lookup.LoaderManager.register_class(*loader_spec)
            assert loader.load_loader((12).__class__) == (loader_spec[0],(*loader_spec[2:0:-1],loader_spec[5]))
    
            # try to find appropriate loader for dict object, a mock of this
            # should have already been imported above 
            assert loader.load_loader(py_object.__class__) == (dict,(load_builtins.create_package_test,b'dict',True))
    
            # remove loader again and undo redirection again. dict should now be
            # processed by create_pickled_dataset
            moc_import_lib.delitem(sys.modules,loader_name)
            del lookup.LoaderManager.__py_types__[None][dict]
            py_obj_type,nopickleloader = loader.load_loader(py_object.__class__)
            assert py_obj_type is object and nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
            
            # check that load_loader prevents redefinition of loaders to be predefined by hickle core
            with pytest.raises(
                RuntimeError,
                match = r"objects\s+defined\s+by\s+hickle\s+core\s+must\s+be"
                        r"\s+registered\s+before\s+first\s+dump\s+or\s+load"
            ):
                py_obj_type,nopickleloader = loader.load_loader(ToBeInLoadersOrNotToBe)
            
            # check that hickle only managed load only loaders without a registered dump function
            # are properly recognized and reported by LoaderManager.load_loader 
            loader.register_class( ToBeInLoadersOrNotToBe, b's.pear', None, load_anyhing,None,True)
            with pytest.raises(
                RuntimeError,
                match = r"objects\s+defined\s+by\s+hickle\s+core\s+must\s+be"
                        r"\s+registered\s+before\s+first\s+dump\s+or\s+load"
            ):
                py_obj_type,nopickleloader = loader.load_loader(ToBeInLoadersOrNotToBe)
            py_obj_type,(create_dataset,base_type,memoise) = loader.load_loader(ToBeInLoadersOrNotToBe,base_type=b's.pear')
            assert py_obj_type is ToBeInLoadersOrNotToBe
            assert create_dataset is lookup.not_dumpable and base_type == b's.pear' and memoise == True
            
            loader.hkl_types_dict.pop(b's.pear',None)
            loader.hkl_container_dict.pop(b's.pear',None)
            loader.types_dict.pop(ToBeInLoadersOrNotToBe,None)
            moc_import_lib.setattr(ToBeInLoadersOrNotToBe,'__module__','hickle.loaders')
    
            # check that load_loaders issues drop warning upon loader definitions for
            # dummy objects defined within hickle package but outside loaders modules
            with pytest.warns(
                RuntimeWarning,
                match = r"ignoring\s+'.+'\s+dummy\s+type\s+not\s+defined\s+by\s+loader\s+module"
            ):
                py_obj_type,nopickleloader = loader.load_loader(ToBeInLoadersOrNotToBe)
                assert py_obj_type is object #ToBeInLoadersOrNotToBe
                assert nopickleloader == (lookup.create_pickled_dataset,b'pickle',True)
    
            # check that loader definitions for dummy objects defined by loaders work as expected
            # by loader module 
            moc_import_lib.setattr(ToBeInLoadersOrNotToBe,'__module__',loader_name)
            py_obj_type,(create_dataset,base_type,memoise) = loader.load_loader(ToBeInLoadersOrNotToBe)
            assert py_obj_type is ToBeInLoadersOrNotToBe and base_type == b'NotHicklable'
            assert create_dataset is not_dumpable
            assert memoise == False
    
            # remove loader_name from list of loaded loaders and check that loader is loaded anew
            # and that values returned for dict object correspond to loader 
            # provided by freshly loaded loader module
            lookup.LoaderManager.__loaded_loaders__.remove(loader_name)
            py_obj_type,(create_dataset,base_type,memoise) = loader.load_loader(py_object.__class__)
            load_builtins_moc = sys.modules.get(loader_name,None)
            assert load_builtins_moc is not None
            loader_spec = load_builtins_moc.class_register[0]
            assert py_obj_type is dict and create_dataset is loader_spec[2]
            assert base_type is loader_spec[1]
            assert memoise == True
            # check that package path is properly resolved if package module
            # for which to find loader for is not found on sys. modules or 
            # its __spec__ attribute is set to None typically on __main__ or
            # builtins and other c modules
            lookup.LoaderManager.__loaded_loaders__.remove(loader_name)
            backup_module = ClassToDump.__module__
            moc_import_lib.setattr(ClassToDump,'__module__',re.sub(r'^\s*hickle\.','',ClassToDump.__module__))
            py_obj_type,(create_dataset,base_type,memoise) = loader.load_loader(ClassToDump)
            assert py_obj_type is object #ClassToDump
            assert create_dataset is lookup.create_pickled_dataset
            assert base_type == b'pickle' and memoise == True
            ClassToDump.__module__ = backup_module
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_hide_collections_loader)
            py_obj_type,(create_dataset,base_type,memoise) = loader.load_loader(collections.OrderedDict)
            moc_import_lib.setattr("hickle.lookup.find_spec",patch_importlib_util_find_spec)
            assert py_obj_type is collections.OrderedDict
            assert create_dataset is sys.modules[loader_name].create_package_test
            assert base_type == b'dict' and memoise == True
            

def test_type_legacy_mro():
    """
    tests type_legacy_mro function which is used in replacement
    for native type.mro function when loading 4.0.0 and 4.0.1 files
    it handles cases where type objects passed to load_loader are
    functions not classes
    """

    # check that for class object type_legacy_mro function returns
    # the mro list provided by type.mro unchanged
    assert lookup.type_legacy_mro(SimpleClass) == type.mro(SimpleClass)

    # check that in case function is passed as type object a tuple with
    # function as single element is returned
    assert lookup.type_legacy_mro(function_to_dump) == (function_to_dump,)


def test_create_pickled_dataset(h5_data,compression_kwargs):
    """
    tests create_pickled_dataset, load_pickled_data function and PickledContainer 
    """
    
    # check if create_pickled_dataset issues SerializedWarning for objects which
    # either do not support copy protocol
    py_object = ClassToDump('hello',1)
    pickled_py_object = pickle.dumps(py_object)
    data_set_name = "greetings"
    with pytest.warns(lookup.SerializedWarning,match = r".*type\s+not\s+understood,\s+data\s+is\s+serialized:.*") as warner:
        h5_node,subitems = lookup.create_pickled_dataset(py_object, h5_data,data_set_name,**compression_kwargs)
        assert isinstance(h5_node,h5py.Dataset) and not subitems and iter(subitems)
        assert bytes(h5_node[()]) == pickled_py_object and h5_node.name.rsplit('/',1)[-1] == data_set_name
        assert lookup.load_pickled_data(h5_node,b'pickle',object) == py_object
        backup_class_to_dump = globals()['ClassToDump']
        backup_class_to_dump = globals().pop('ClassToDump',None)
        recovered = lookup.load_pickled_data(h5_node,b'pickle',object)
        assert isinstance(recovered,lookup.RecoveredDataset)
        assert bytes(recovered) == pickled_py_object
        globals()['ClassToDump'] = backup_class_to_dump
    
    
def test__DictItemContainer():
    """
    tests _DictItemContainer class which represent dict_item group 
    used by version 4.0.0 files to represent values of dictionary key
    """
    container = lookup._DictItemContainer({},b'dict_item',lookup._DictItem)
    my_bike_lock = (1,2,3,4)
    container.append('my_bike_lock',my_bike_lock,{})
    assert container.convert() is my_bike_lock

    
#@pytest.mark.no_compression
def test__moc_numpy_array_object_lambda():
    """
    test the _moc_numpy_array_object_lambda function
    which mimics the effect of lambda function created
    py pickle when expanding pickle `'type'` string set
    for numpy arrays containing a single object not expandable
    into a list. Mocking is necessary from Python 3.8.X on
    as it seems in Python 3.8 and onward trying to pickle
    a lambda now causes a TypeError whilst it seems to be silently
    accepted in Python < 3.8
    """
    data = ['hello','world']
    assert lookup._moc_numpy_array_object_lambda(data) == data[0]
    
#@pytest.mark.no_compression
def test_fix_lambda_obj_type():
    """
    test _moc_numpy_array_object_lambda function it self. When invoked
    it should return the first element of the passed list
    """
    assert lookup.fix_lambda_obj_type(None) is object
    picklestring = pickle.dumps(SimpleClass)
    assert lookup.fix_lambda_obj_type(picklestring) is SimpleClass
    with pytest.warns(lookup.MockedLambdaWarning):
        assert lookup.fix_lambda_obj_type('') is lookup._moc_numpy_array_object_lambda

def test_ReferenceManager_get_root(h5_data):
    """
    tests the static ReferenceManager._get_root method
    """

    # create an artificial 'hickle_types_table' with some entries
    # and link their h5py.Reference objects to the 'type' attributes 
    # of some data such that ReferenceManager._get_root can resolve
    # h5_data root_group independent which node it was passed
    
    root_group = h5_data['/root_group']
    data_group = root_group.create_group('data')
    content = data_group.create_dataset('mydata',data=12)
    type_table = root_group.create_group('hickle_types_table')
    int_pickle_string = bytearray(pickle.dumps(int))
    int_np_entry = np.array(int_pickle_string,copy=False)
    int_np_entry.dtype = 'S1'
    int_entry = type_table.create_dataset(str(len(type_table)),data = int_np_entry,shape =(1,int_np_entry.size))
    int_base_type = b'int'
    int_base_type = type_table.create_dataset(int_base_type,shape=None,dtype="S1")
    int_entry.attrs['base_type'] = int_base_type.ref
    content.attrs['type'] = int_entry.ref
    # try to resolve root_group from various kinds of nodes including 
    # root_group it self.
    assert lookup.ReferenceManager.get_root(content).id == root_group.id
    assert lookup.ReferenceManager.get_root(root_group).id == root_group.id
    assert lookup.ReferenceManager.get_root(data_group).id == data_group.id

    # check fallbacks to passe in group or file in case resolution via
    # 'type' attribute reference fails
    list_group = data_group.create_group('somelist')
    some_list_item = list_group.create_dataset('0',data=13)
    assert lookup.ReferenceManager.get_root(some_list_item).id == some_list_item.file.id
    assert lookup.ReferenceManager.get_root(list_group).id == list_group.id

    # test indirect resolution through 'type' reference of parent group
    # which should have an already properly assigned 'type' attribute
    # unless reading hickle 4.0.X file or referred to 'hickle_types_table' entry
    # is missing. In both cases file shall be returned as fallback
    list_pickle_string = bytearray(pickle.dumps(list))
    list_np_entry = np.array(list_pickle_string,copy = False)
    list_np_entry.dtype = 'S1'
    list_entry = type_table.create_dataset(str(len(type_table)),data = list_np_entry,shape=(1,list_np_entry.size))
    list_base_type = b'list'
    list_base_type = type_table.create_dataset(list_base_type,shape=None,dtype="S1")
    list_entry.attrs['base_type'] = list_base_type.ref
    list_group.attrs['type'] = list_pickle_string
    assert lookup.ReferenceManager.get_root(some_list_item).id == root_group.file.id
    list_group.attrs['type'] = list_entry.ref
    assert lookup.ReferenceManager.get_root(some_list_item).id == root_group.id
    for_later_use = list_entry.ref
    list_entry = None
    del type_table[str(len(type_table)-2)]
    assert lookup.ReferenceManager.get_root(some_list_item).id == root_group.file.id
    assert lookup.ReferenceManager.get_root(some_list_item).id == root_group.file.id
    
        

class not_a_surviver():
    """does not survive pickle.dumps"""

def test_ReferenceManager(h5_data):
    """
    test for creation of ReferenceManager object (__init__)
    to be run before testing ReferenceManager.create_manager
    """
    
    reference_manager = lookup.ReferenceManager(h5_data)
    type_table = h5_data['hickle_types_table']
    assert isinstance(type_table,h5py.Group)
    reference_manager = lookup.ReferenceManager(h5_data)
    assert reference_manager._py_obj_type_table.id == type_table.id
    false_root = h5_data.file.create_group('false_root')
    false_root.create_dataset('hickle_types_table',data=12)
    with pytest.raises(lookup.ReferenceError):
        reference_manager = lookup.ReferenceManager(false_root)
    int_pickle_string = bytearray(pickle.dumps(int))
    int_np_entry = np.array(int_pickle_string,copy=False)
    int_np_entry.dtype = 'S1'
    int_entry = type_table.create_dataset(str(len(type_table)),data = int_np_entry,shape =(1,int_np_entry.size))
    int_base_type = b'int'
    int_base_type = type_table.create_dataset(int_base_type,shape=None,dtype="S1")
    int_entry.attrs['base_type'] = int_base_type.ref
    list_pickle_string = bytearray(pickle.dumps(list))
    list_np_entry = np.array(list_pickle_string,copy = False)
    list_np_entry.dtype = 'S1'
    list_entry = type_table.create_dataset(str(len(type_table)),data = list_np_entry,shape=(1,list_np_entry.size))
    list_base_type = b'list'
    list_base_type = type_table.create_dataset(list_base_type,shape=None,dtype="S1")
    list_entry.attrs['base_type'] = list_base_type.ref

    missing_pickle_string = bytearray(pickle.dumps(not_a_surviver))
    missing_np_entry = np.array(missing_pickle_string,copy = False)
    missing_np_entry.dtype = 'S1'
    missing_entry = type_table.create_dataset(str(len(type_table)),data = missing_np_entry,shape=(1,missing_np_entry.size))
    missing_base_type = b'lost'
    missing_base_type = type_table.create_dataset(missing_base_type,shape=None,dtype="S1")
    missing_entry.attrs['base_type'] = missing_base_type.ref
    hide_not_a_surviver = globals().pop('not_a_surviver',None)
    reference_manager = lookup.ReferenceManager(h5_data)
    globals()['not_a_surviver'] = hide_not_a_surviver
    assert reference_manager._py_obj_type_link[id(int)] == int_entry
    assert reference_manager._py_obj_type_link[int_entry.id] == (int,b'int')
    assert reference_manager._base_type_link[b'int'] == int_base_type
    assert reference_manager._base_type_link[int_base_type.id] == b'int'
    assert reference_manager._py_obj_type_link[id(list)] == list_entry
    assert reference_manager._py_obj_type_link[list_entry.id] == (list,b'list')
    assert reference_manager._base_type_link[b'list'] == list_base_type
    assert reference_manager._base_type_link[list_base_type.id] == b'list'
    assert reference_manager._base_type_link[b'lost'] == missing_base_type
    assert reference_manager._base_type_link[missing_base_type.id] == b'lost'
    assert reference_manager._py_obj_type_link[missing_entry.id] == (lookup.AttemptRecoverCustom,'!recover!',b'lost')
    backup_attr = list_entry.attrs['base_type']
    list_entry.attrs.pop('base_type',None)
    with pytest.raises(lookup.ReferenceError):
        reference_manager = lookup.ReferenceManager(h5_data)
    list_entry.attrs['base_type']=b'list'
    with pytest.raises(lookup.ReferenceError):
        reference_manager = lookup.ReferenceManager(h5_data)
    stale_ref_entry = type_table.create_dataset("stale",shape=None,dtype = 'S1')
    list_entry.attrs['base_type']=stale_ref_entry.ref
    type_table.pop("stale",None)
    stale_ref_entry = None
    with pytest.raises(lookup.ReferenceError):
        reference_manager = lookup.ReferenceManager(h5_data)
    list_entry.attrs['base_type']=backup_attr
    
    old_hickle_file_root = h5_data.file.create_group('old_root')
    h5_data.file.flush()
    base_name,ext = h5_data.file.filename.rsplit('.',1)
    file_name = "{}_ro.{}".format(base_name,ext)
    data_name = h5_data.name
    data_file_name = h5_data.file.filename
    #################### NOTE #############################
    # h5_data ficture is invalidated by the following line
    # as well as all Groups and Datasets read from it
    # manually reopen it and do not forget to close at
    # end of iths test
    h5_data.file.close()
    shutil.copy(data_file_name,file_name)
    read_only_handle = h5py.File(file_name,'r')
    h5_read_data = read_only_handle[data_name]
    h5_read_old = read_only_handle['old_root']
    reference_manager = lookup.ReferenceManager(h5_read_old)
    assert isinstance(reference_manager._overlay,weakref.finalize)
    overlay_file = reference_manager._py_obj_type_table.file
    assert overlay_file.mode == 'r+' and overlay_file.driver == 'core'
    assert overlay_file.id != read_only_handle.id
    reference_manager = lookup.ReferenceManager(h5_read_data)
    
    
    read_only_handle.close()
    class SubReferenceManager(lookup.ReferenceManager):
        __managers__ = ()
    assert SubReferenceManager.__managers__ is lookup.ReferenceManager.__managers__

    with pytest.raises(TypeError):
        invalid_instance = lookup.BaseManager()

    class OtherManager(lookup.BaseManager):
        pass

    with pytest.raises(NotImplementedError):
        with OtherManager() as invalid_manager:
            pass

    

    
def test_ReferenceManager_drop_manager(h5_data):
    """
    test static ReferenceManager._drop_table method
    """
    reference_manager = lookup.ReferenceManager(h5_data)
    lookup.ReferenceManager.__managers__[h5_data.file.id] = (reference_manager,h5_data)
    some_other_file = h5py.File('someother.hdf5','w')
    some_other_root = some_other_file.create_group('root')
    lookup.ReferenceManager._drop_manager(some_other_root.file.id)
    lookup.ReferenceManager.__managers__[some_other_file.file.id] = (lookup.ReferenceManager(some_other_root),some_other_root)
    assert lookup.ReferenceManager.__managers__.get(h5_data.file.id,None) == (reference_manager,h5_data)
    lookup.ReferenceManager._drop_manager(h5_data.file.id)
    assert lookup.ReferenceManager.__managers__.get(h5_data.file.id,None) is None
    lookup.ReferenceManager._drop_manager(some_other_root.file.id)
    assert not lookup.ReferenceManager.__managers__
    some_other_file.close()
    
def test_ReferenceManager_create_manager(h5_data):
    """
    test public static ReferenceManager.create_manager function
    """
    second_tree = h5_data.file.create_group('seondary_root')
    h5_data_table = lookup.ReferenceManager.create_manager(h5_data)
    assert lookup.ReferenceManager.__managers__[h5_data.file.id][0] is h5_data_table
    with pytest.raises(lookup.LookupError):
        second_table = lookup.ReferenceManager.create_manager(second_tree)
    lookup.ReferenceManager._drop_manager(h5_data.file.id)

def test_ReferenceManager_context(h5_data):
    """
    test use of ReferenceManager as context manager
    """
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        assert lookup.ReferenceManager.__managers__[h5_data.file.id][0] is memo
    assert memo._py_obj_type_table is None
    with pytest.raises(RuntimeError):
        with memo as memo2:
            pass
    memo.__exit__(None,None,None)
    old_hickle_file_root = h5_data.file.create_group('old_root')
    h5_data.file.flush()
    base_name,ext = h5_data.file.filename.rsplit('.',1)
    file_name = "{}_ro.{}".format(base_name,ext)
    data_file_name = h5_data.file.filename
    data_name = old_hickle_file_root.name
    #################### NOTE #############################
    # h5_data ficture is invalidated by the following line
    # as well as all Groups and Datasets read from it
    # manually reopen it and do not forget to close at
    # end of iths test
    h5_data.file.close()
    shutil.copy(data_file_name,file_name)
    read_only_handle = h5py.File(file_name,'r')
    h5_read_data = read_only_handle[data_name]
    with lookup.ReferenceManager.create_manager(h5_read_data) as memo:
        assert isinstance(memo._overlay,weakref.finalize)
    assert memo._overlay is None
    read_only_handle.close()
        
def test_ReferenceManager_store_type(h5_data,compression_kwargs):
    """
    test ReferenceManager.store_type method which sets 'type' attribute
    reference to appropriate py_obj_type entry within 'hickle_types_table'
    """
    h_node = h5_data.create_group('some_list')
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        memo.store_type(h_node,object,None,**compression_kwargs)
        assert len(memo._py_obj_type_table) == 0 and not memo._py_obj_type_link and not memo._base_type_link
        with pytest.raises(lookup.LookupError):
            memo.store_type(h_node,list,None,**compression_kwargs)
        with pytest.raises(ValueError):
            memo.store_type(h_node,list,b'',**compression_kwargs)
        memo.store_type(h_node,list,b'list',**compression_kwargs)
        assert isinstance(h_node.attrs['type'],h5py.Reference)
        type_table_entry = h5_data.file[h_node.attrs['type']]
        assert pickle.loads(type_table_entry[()]) is list
        assert isinstance(type_table_entry.attrs['base_type'],h5py.Reference)
        assert h5_data.file[type_table_entry.attrs['base_type']].name.rsplit('/',1)[-1].encode('ascii') == b'list'
    
@pytest.mark.no_compression
def test_ReferenceManager_get_manager(h5_data):
    h_node = h5_data.create_group('some_list')
    item_data = np.array(memoryview(b'hallo welt lore grueszet dich ipsum aus der lore von ipsum gelort in ipsum'),copy=False)
    item_data.dtype = 'S1'
    h_item = h_node.create_dataset('0',data=item_data,shape=(1,item_data.size))
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        memo.store_type(h_node,list,b'list')
        memo.store_type(h_item,bytes,b'bytes')
        assert lookup.ReferenceManager.get_manager(h_item) == memo

        backup_manager = lookup.ReferenceManager.__managers__.pop(h5_data.file.id,None)
        assert backup_manager is not None
        with pytest.raises(lookup.ReferenceError):
            manager = lookup.ReferenceManager.get_manager(h_item)
        lookup.ReferenceManager.__managers__[h5_data.file.id] = backup_manager
    with pytest.raises(lookup.ReferenceError):
        manager = lookup.ReferenceManager.get_manager(h_item)

@pytest.mark.no_compression
def test_ReferenceManager_resolve_type(h5_data):
    """
    test ReferenceManager.reslove_type method which tries to resolve
    content of 'type' attribute of passed in node and return appropriate pair of
    py_obj_type,base_type and a boolean flag indicating whether node represents
    a h5py.Group or h5py.Reference both of which are to be handled by PyContainer
    objects.
    """
    invalid_pickle_and_ref = h5_data.create_group('invalid_pickle_and_ref')
    pickled_data = h5_data.create_dataset('pickled_data',data = bytearray())
    shared_ref = h5_data.create_dataset('shared_ref',data = pickled_data.ref,dtype = h5py.ref_dtype)
    old_style_typed = h5_data.create_dataset('old_style_typed',data = 12)
    old_style_typed.attrs['type'] = np.array(pickle.dumps(int))
    old_style_typed.attrs['base_type'] = b'int'
    broken_old_style = h5_data.create_dataset('broken_old_style',data = 12)
    broken_old_style.attrs['type'] = 12
    broken_old_style.attrs['base_type'] = b'int'
    new_style_typed = h5_data.create_dataset('new_style_typed',data = 12)
    stale_new_style = h5_data.create_dataset('stale_new_style',data = 12)
    new_style_typed_no_link = h5_data.create_dataset('new_style_typed_no_link',data = 12.5)
    has_not_recoverable_type = h5_data.create_dataset('no_recoverable_type',data = 42.56)
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        with pytest.raises(lookup.ReferenceError):
            memo.resolve_type(invalid_pickle_and_ref)
        assert memo.resolve_type(pickled_data) == (object,b'pickle',False)
        assert memo.resolve_type(shared_ref) == (lookup.NodeReference,b'!node-reference!',True)
        assert memo.resolve_type(old_style_typed) in ((int,b'int',False),(int,'int',False))
        with pytest.raises(lookup.ReferenceError):
            info = memo.resolve_type(broken_old_style)
        memo.store_type(new_style_typed,int,b'int')
        entry_id = len(memo._py_obj_type_table)
        memo.store_type(stale_new_style,list,b'list')
        assert pickle.loads(memo._py_obj_type_table[str(entry_id)][()]) is list
        stale_list_base = memo._py_obj_type_table['list'].ref
        # remove py_obj_type entry and base_type_entry for list entry
        # while h5py 2 raises a value error if not active link exists for a
        # dataset h5py 3 returns an anonymous group when resolving a stale 
        # reference to it if any body still holds a strong reference to its
        # h5py.Dataset or h5py.Group object. Therefore drop all references to
        # the removed entries to simulate that somebody has removed them from
        # a hickle file before it was passed to hickle.load for restoring its
        # content.
        memo._py_obj_type_link.pop(memo._py_obj_type_table[str(entry_id)].id,None)
        memo._py_obj_type_link.pop(id(list),None)
        memo._base_type_link.pop(memo._py_obj_type_table['list'].id,None)
        memo._base_type_link.pop(b'list',None)
        del memo._py_obj_type_table[str(entry_id)]
        del memo._py_obj_type_table['list']
        memo._py_obj_type_table.file.flush()
        with pytest.raises(lookup.ReferenceError):
            memo.resolve_type(stale_new_style)
        entry_id = len(memo._py_obj_type_table)
        memo.store_type(new_style_typed_no_link,float,b'float')
        float_entry = memo._py_obj_type_table[str(entry_id)]
        assert pickle.loads(float_entry[()]) is float
        float_base = float_entry.attrs['base_type']
        # remove float entry and clear all references to it see above 
        del memo._py_obj_type_link[float_entry.id]
        del memo._py_obj_type_link[id(float)]
        del float_entry.attrs['base_type']
        memo._py_obj_type_table.file.flush()
        assert memo.resolve_type(new_style_typed_no_link) in ((float,b'pickle',False),(float,'pickle',False))
        del memo._py_obj_type_link[float_entry.id]
        del memo._py_obj_type_link[id(float)]
        # create stale reference to not existing base_type entry
        memo._py_obj_type_table.create_dataset('list',shape=None,dtype='S1')
        float_entry.attrs['base_type'] = memo._py_obj_type_table['list'].ref
        memo._py_obj_type_table.pop('list',None)
        memo._py_obj_type_table.file.flush()
        with pytest.raises(lookup.ReferenceError):
            info = memo.resolve_type(new_style_typed_no_link)
        memo._py_obj_type_link.pop(float_entry.id,None)
        memo._py_obj_type_link.pop(id(float),None)
        del memo._base_type_link[memo._py_obj_type_table[float_base].id]
        del memo._base_type_link[b'float']
        float_entry.attrs['base_type'] = float_base
        memo._py_obj_type_table.file.flush()
        assert memo.resolve_type(new_style_typed_no_link) in ((float,b'float',False),(float,'float',False))
        
        assert memo.resolve_type(new_style_typed_no_link) 
        memo.store_type(has_not_recoverable_type,not_a_surviver,b'lost')
        del memo._py_obj_type_link[memo._py_obj_type_link[id(not_a_surviver)].id]
        del memo._py_obj_type_link[id(not_a_surviver)]
        hide_not_a_surviver = globals().pop('not_a_surviver',None)
        assert memo.resolve_type(has_not_recoverable_type) == (lookup.AttemptRecoverCustom,b'!recover!',False)
        assert memo.resolve_type(has_not_recoverable_type,base_type_type=2) == (lookup.AttemptRecoverCustom,b'lost',False)
        globals()['not_a_surviver'] = hide_not_a_surviver
        has_not_recoverable_type.attrs['type'] = np.array(pickle.dumps(not_a_surviver))
        has_not_recoverable_type.attrs['base_type'] = b'lost'
        hide_not_a_surviver = globals().pop('not_a_surviver',None)
        assert memo.resolve_type(has_not_recoverable_type) == (lookup.AttemptRecoverCustom,b'!recover!',False)
        assert memo.resolve_type(has_not_recoverable_type,base_type_type=2) in ((lookup.AttemptRecoverCustom,b'lost',False),(lookup.AttemptRecoverCustom,'lost',False))
        globals()['not_a_surviver'] = hide_not_a_surviver
    
    
def test_ExpandReferenceContainer(h5_data):
    """
    test ExpandReferenceContainer which resolves object link stored as h5py.Reference
    type dataset
    """
    expected_data = np.random.randint(-13,13,12)
    referred_data = h5_data.create_dataset('referred_data',data = expected_data)
    referring_node = h5_data.create_dataset('referring_node',data = referred_data.ref,dtype = h5py.ref_dtype)
    h5_data.file.flush()
    sub_container = lookup.ExpandReferenceContainer(referring_node.attrs,b'!node-reference!',lookup.NodeReference)
    content = None
    for name,subitem in sub_container.filter(referring_node):
        assert name == 'referred_data' and subitem.id == referred_data.id
        content = np.array(subitem[()])
        sub_container.append(name,content,subitem.attrs)
    assert np.all(sub_container.convert()==expected_data)
    referring_node = h5_data.create_dataset('stale_reference',shape=(),dtype=h5py.ref_dtype)
    sub_container = lookup.ExpandReferenceContainer(referring_node.attrs,b'!node-reference!',lookup.NodeReference)
    with pytest.raises(lookup.ReferenceError):
        for name,subitem in sub_container.filter(referring_node):
            content = np.array(subitem[()])
            sub_container.append(name,content,subitem.attrs)

@pytest.mark.no_compression
def test_recover_custom_data(h5_data):
    array_to_recover = np.random.random_sample([4,2])
    with lookup.ReferenceManager.create_manager(h5_data) as memo:
        dataset_to_recover = h5_data.create_dataset('to_recover',data=array_to_recover)
        dataset_to_recover.attrs['world'] = 2
        memo.store_type(dataset_to_recover,ClassToDump,b'myclass')
        group_to_recover = h5_data.create_group('need_recover')
        memo.store_type(group_to_recover,ClassToDump,b'myclass')
        backup_class_to_dump = globals().pop('ClassToDump',None)
        memo._py_obj_type_link.pop(id('ClassToDump'),None)
        memo._base_type_link.pop(b'myclass')
        type_entry = memo._py_obj_type_table[dataset_to_recover.attrs['type']]
        memo._py_obj_type_link.pop(type_entry.id,None)
        py_obj_type,base_type,is_group = memo.resolve_type(dataset_to_recover)
        assert issubclass(py_obj_type,lookup.AttemptRecoverCustom) and base_type == b'!recover!'
        with pytest.warns(lookup.DataRecoveredWarning):
            recovered = lookup.recover_custom_dataset(dataset_to_recover,base_type,py_obj_type) 
        assert recovered.dtype == array_to_recover.dtype and np.all(recovered == array_to_recover)
        assert recovered.attrs == {'base_type':b'myclass','world':2}
        assert not is_group
        type_entry = memo._py_obj_type_table[group_to_recover.attrs['type']]
        memo._py_obj_type_link.pop(type_entry.id,None)
        some_int=group_to_recover.create_dataset('some_int',data=42)
        some_float=group_to_recover.create_dataset('some_float',data=42.0)
        group_to_recover.attrs['so'] = 'long'
        group_to_recover.attrs['and'] = 'thanks'
        some_float.attrs['for'] = 'all'
        some_float.attrs['the'] = 'fish'
        py_obj_type,base_type,is_group = memo.resolve_type(group_to_recover)
        assert issubclass(py_obj_type,lookup.AttemptRecoverCustom) and base_type == b'!recover!'
        assert is_group
        recover_container = lookup.RecoverGroupContainer(group_to_recover.attrs,base_type,py_obj_type)
        with pytest.warns(lookup.DataRecoveredWarning):
            for name,item in recover_container.filter(group_to_recover):
                recover_container.append(name,item[()],item.attrs)
        recover_container.append('some_other',recovered,recovered.attrs)
        recovered_group = recover_container.convert()
        assert isinstance(recovered_group,dict)
        assert some_float[()] == recovered_group['some_float'][0] and some_float.attrs == recovered_group['some_float'][1]
        assert some_int[()] == recovered_group['some_int'][0] and some_int.attrs == recovered_group['some_int'][1]
        assert recovered_group['some_other'] is recovered
        assert recovered_group.attrs['base_type'] == memo.resolve_type(group_to_recover,base_type_type=2)[1]
        assert len(recovered_group.attrs) == 3
        assert recovered_group.attrs['so'] == 'long' and recovered_group.attrs['and'] == 'thanks'
        globals()['ClassToDump'] = backup_class_to_dump 

if __name__ == "__main__":
    from _pytest.monkeypatch import monkeypatch
    from _pytest.fixtures import FixtureRequest
    from hickle.tests.conftest import compression_kwargs

    for h5_root in h5_data(FixtureRequest(test_create_pickled_dataset)):
        test_AttemptRecoverCustom_classes(h5_data)
    for table in loader_table():
        test_LoaderManager_register_class(table)
    for table in loader_table():
        test_LoaderManager_register_class_exclude(table)
    for table,h5_root in (
        (tab,root)
        for tab in loader_table()
        for root in h5_data(FixtureRequest(test_LoaderManager))
    ):
        test_LoaderManager(table,h5_root)
    for h5_root in h5_data(FixtureRequest(test_LoaderManager_drop_manager)):
        test_LoaderManager_drop_manager(h5_root)
    for h5_root in h5_data(FixtureRequest(test_LoaderManager_create_manager)):
        test_LoaderManager_create_manager(h5_root)
    for h5_root in h5_data(FixtureRequest(test_LoaderManager_context)):
        test_LoaderManager_context(h5_root)
    for table,h5_root,monkey in (
        (tab,root,mpatch)
        for tab in loader_table()
        for root in h5_data(FixtureRequest(test_LoaderManager_load_loader))
        for mpatch in monkeypatch()
    ):
            test_LoaderManager_load_loader(table,h5_root,monkey)
    test_type_legacy_mro()
    for h5_root,keywords in (
        ( h5_data(request),compression_kwargs(request) )
        for request in (FixtureRequest(test_create_pickled_dataset),)
    ):
        test_create_pickled_dataset(h5_root,keywords)
    test__DictItemContainer()
    test__moc_numpy_array_object_lambda()
    test_fix_lambda_obj_type()
    test_fix_lambda_obj_type()
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_get_root)):
        test_ReferenceManager_get_root(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager)):
        test_ReferenceManager(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_drop_manager)):
        test_ReferenceManager_drop_manager(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_create_manager)):
        test_ReferenceManager_create_manager(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_context)):
        test_ReferenceManager_context(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_get_manager)):
        test_ReferenceManager_get_manager(h5_root)
    for h5_root,compression_kwargs in (
            h5_data(FixtureRequest(test_ReferenceManager_store_type))
    ):
        test_ReferenceManager_store_type(h5_root,compression_kwargs)
    for h5_root in h5_data(FixtureRequest(test_ReferenceManager_resolve_type)):
        test_ReferenceManager_resolve_type(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ExpandReferenceContainer)):
        test_ExpandReferenceContainer(h5_root)
    for h5_root in h5_data(FixtureRequest(test_ExpandReferenceContainer)):
        test_recover_custom_data(h5_data)


    
    




# %% IMPORTS
# Package imports
from astropy.coordinates import Angle, SkyCoord
from astropy.constants import Constant
from astropy.table import Table
from astropy.time import Time
from astropy.units import Quantity
import numpy as np

# hickle imports
from hickle.helpers import no_compression,load_str_list_attr_ascii,load_str_attr_ascii


# %% FUNCTION DEFINITIONS
def create_astropy_quantity(py_obj, h_group, name, **kwargs):
    """
    dumps an astropy quantity

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset representing astropy quantity and empty subitems
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **no_compression(kwargs))
    d.attrs['unit'] = py_obj.unit.to_string().encode('ascii')
    return d,()


def create_astropy_angle(py_obj, h_group, name, **kwargs):
    """
    dumps an astropy angle

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset representing astropy angle and empty subitems
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **no_compression(kwargs))
    d.attrs['unit'] = py_obj.unit.to_string().encode('ascii')
    return d,()


def create_astropy_skycoord(py_obj, h_group, name, **kwargs):
    """
    dumps an astropy SkyCoord object

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset representing astorpy SkyCoord and empty subitems
    """

    lon = py_obj.data.lon.value
    lat = py_obj.data.lat.value
    dd = np.stack((lon, lat), axis=-1)

    d = h_group.create_dataset(name, data=dd, dtype='float64', **kwargs)
    lon_unit = py_obj.data.lon.unit.to_string().encode('ascii')
    lat_unit = py_obj.data.lat.unit.to_string().encode('ascii')
    d.attrs['lon_unit'] = lon_unit
    d.attrs['lat_unit'] = lat_unit
    return d,()


def create_astropy_time(py_obj, h_group, name, **kwargs):
    """ dumps an astropy Time object

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset representing astropy time and empty subitems
    """

    # Need to catch string times, e.g. 1999-01-01T00:00:00.123
    # Must be encoded into bytes. 
    if 'str' in py_obj.value.dtype.name:
        bytes_dtype_str = py_obj.value.dtype.str.replace('<U', '|S')
        d = h_group.create_dataset(
            name,
            data = np.array(py_obj.value.astype(bytes_dtype_str)),
            **kwargs
        )
    else:
        d = h_group.create_dataset(name, data=py_obj.value, dtype=py_obj.value.dtype)
    d.attrs['np_dtype'] = py_obj.value.dtype.str.encode('ascii')

    d.attrs['format'] = str(py_obj.format).encode('ascii')
    d.attrs['scale'] = str(py_obj.scale).encode('ascii')

    return d,()


def create_astropy_constant(py_obj, h_group, name, **kwargs):
    """ dumps an astropy Constant

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset representing astropy constant and empty subitems list
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **no_compression(kwargs))
    d.attrs["unit"] = py_obj.unit.to_string().encode('ascii')
    d.attrs["abbrev"] = py_obj.abbrev.encode('ascii')
    d.attrs["name"] = py_obj.name.encode('ascii')
    d.attrs["reference"] = py_obj.reference.encode('ascii')
    d.attrs["uncertainty"] = py_obj.uncertainty

    if py_obj.system:
        d.attrs["system"] = py_obj.system.encode('ascii')
    return d,()


def create_astropy_table(py_obj, h_group, name, **kwargs):
    """ Dump an astropy Table

    Parameters
    ----------
    py_obj:
        python object to dump; should be a python type

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function
  
    Returns
    -------
    tuple containing h5py.Dataset representing astropy table and empty subitems list
    """
    data = py_obj.as_array()
    d = h_group.create_dataset(name, data=data, dtype=data.dtype, **kwargs)

    colnames = [cn.encode('ascii') for cn in py_obj.colnames]
    d.attrs['colnames'] = colnames
    for key, value in py_obj.meta.items():
        d.attrs[key] = value
    return d,()


def load_astropy_quantity_dataset(h_node,base_type,py_obj_type):
    """
    loads astropy Quantity

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
    resulting py_obj_type
    """
    unit = h_node.attrs["unit"]
    return py_obj_type(h_node[()], unit)


def load_astropy_time_dataset(h_node,base_type,py_obj_type):
    """
    loads astropy time

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
    resulting py_obj_type
    """
    fmt = load_str_attr_ascii(h_node.attrs,"format")
    scale = load_str_attr_ascii(h_node.attrs,"scale")
    dtype = h_node.attrs.get('np_dtype','')
    if dtype:
        dtype = np.dtype(dtype)
        if 'str' in dtype.name:
            bytes_dtype_str = dtype.str.replace('|S', '<U')
            time_data = np.array(h_node[()]).astype(bytes_dtype_str)
            return py_obj_type(time_data, format=fmt, scale=scale)
        return py_obj_type(np.array(h_node[()],dtype=dtype), format=fmt, scale=scale)
    return py_obj_type(np.array(h_node[()],dtype = h_node.dtype), format=fmt, scale=scale)


def load_astropy_angle_dataset(h_node,base_type,py_obj_type):
    """
    loads astropy angle

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
        resulting py_obj_type
    """
    unit = h_node.attrs["unit"]
    q = py_obj_type(h_node[()], unit)
    return q


def load_astropy_skycoord_dataset(h_node,base_type,py_obj_type):
    """
    loads astropy SkyCoord

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
    resulting py_obj_type
    """
    data = h_node[()]
    lon_unit = load_str_attr_ascii(h_node.attrs,"lon_unit")
    lat_unit = load_str_attr_ascii(h_node.attrs,"lat_unit")
    q = py_obj_type(data[..., 0], data[..., 1], unit=(lon_unit, lat_unit))
    return q


def load_astropy_constant_dataset(h_node,base_type,py_obj_type):
    """
    loads astropy constant

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
    resulting py_obj_type
    """
    unit = h_node.attrs["unit"]
    abbrev = h_node.attrs["abbrev"]
    name = h_node.attrs["name"]
    ref = h_node.attrs["reference"]
    unc = h_node.attrs["uncertainty"]

    system = None
    if "system" in h_node.attrs.keys():
        system = h_node.attrs["system"]

    c = py_obj_type(abbrev, name, h_node[()], unit, unc, ref, system)
    return c


def load_astropy_table(h_node,base_type,py_obj_type):
    """
    loads astropy table

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type :
        final type of restored dtype

    Returns
    -------
    resulting py_obj_type
    """

    colnames = load_str_list_attr_ascii(h_node.attrs,"colnames")

    t = py_obj_type(
        h_node[()],
        names = colnames,
        meta = {
            metakey:metavalue 
            for metakey,metavalue in h_node.attrs.items()
            if metakey not in {'type','base_type','colnames'}
        }
    )
    return t

# %% REGISTERS
class_register = [
    [Quantity, b'astropy_quantity', create_astropy_quantity, load_astropy_quantity_dataset],
    [Time, b'astropy_time', create_astropy_time, load_astropy_time_dataset],
    [Angle, b'astropy_angle', create_astropy_angle, load_astropy_angle_dataset],
    [SkyCoord, b'astropy_skycoord', create_astropy_skycoord, load_astropy_skycoord_dataset],
    [Constant, b'astropy_constant', create_astropy_constant, load_astropy_constant_dataset],
    [Table, b'astropy_table',  create_astropy_table, load_astropy_table]
]

exclude_register = []



# encoding: utf-8
"""
# load_numpy.py

Utilities and dump / load handlers for handling numpy and scipy arrays

"""

# %% IMPORTS
# Package imports
import numpy as np
import types

# hickle imports
from hickle.loaders.load_builtins import create_listlike_dataset,ListLikeContainer
from hickle.helpers import PyContainer,no_compression


# %% FUNCTION DEFINITIONS

def create_np_scalar_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an numpy.dtype object to h5py file

    Parameters
    ----------
    py_obj (numpy.scalar):
        python object to dump; should be a numpy scalar, e.g.  numpy.float16(1)

    h_group (h5.File.group):
        group to dump data into.

    name (str):
         the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset and empty list of subitems
    """

    d = h_group.create_dataset(name, data=py_obj, **no_compression(kwargs))

    d.attrs["np_dtype"] = py_obj.dtype.str.encode("ascii")
    return d,()


def create_np_dtype(py_obj, h_group, name, **kwargs):
    """ dumps an numpy dtype object to h5py file

    Parameters
    ----------
    py_obj (numpy.dtype):
        python object to dump; should be a numpy dtype, e.g.  numpy.float16

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Dataset and empty list of subitems
    """
    d = h_group.create_dataset(name, data=bytearray(py_obj.str,"ascii"), **kwargs)
    return d,()


def create_np_array_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an ndarray object to h5py file

    Parameters
    ----------
    py_obj (numpy.ndarray):
        python object to dump; should be a numpy.ndarray or numpy.ma.array (masked)

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset or group

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Datset and empty list of subitems or h5py.Group
    and iterable of subitems
    """

    # Obtain dtype of py_obj
    dtype = py_obj.dtype

    # Check if py_obj contains strings
    if "str" in dtype.name:
        if py_obj.ndim < 1:
            # convert string to utf8 encoded bytearray
            string_data = bytearray(py_obj.item(),"utf8") if 'bytes' not in dtype.name else memoryview(py_obj.item())
            string_data = np.array(string_data,copy = False)
            string_data.dtype = 'S1'
            h_node = h_group.create_dataset(name,data = string_data,shape=(1,string_data.size),**kwargs)
            sub_items = ()
        else:
            # store content as list of strings
            h_node,sub_items = create_listlike_dataset(py_obj.tolist(), h_group, name, **kwargs)
    elif dtype.name == 'object':
        # If so, convert py_obj to list
        py_obj = py_obj.tolist()

        # Check if py_obj is a list
        if isinstance(py_obj, list):
            # If so, dump py_obj into the current group
            h_node,sub_items = create_listlike_dataset(py_obj, h_group, name, **kwargs)
        else:
            # If not, create a new group and dump py_obj into that
            h_node = h_group.create_group(name)
            sub_items = ("data",py_obj,{},kwargs),
    else:
        h_node = h_group.create_dataset(
            name, data=py_obj, **( no_compression(kwargs) if "bytes" in dtype.name else kwargs )
        )
        sub_items = ()
    h_node.attrs['np_dtype'] = dtype.str.encode('ascii')
    return h_node,sub_items

def create_np_masked_array_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an numpy.ma.core.MaskedArray object to h5py file

    Parameters
    ----------
    py_obj (numpy.ma.array):
        python object to dump; should be a numpy.ndarray or numpy.ma.array (masked)

    h_group (h5.File.group):
        group to dump data into.

    name (str):
        the name of the resulting dataset or group

    kwargs (dict):
        keyword arguments to be passed to create_dataset function

    Returns
    -------
    tuple containing h5py.Group and subitems list representing masked array contents:

    """

    # Obtain dtype of py_obj

    h_node = h_group.create_group(name)
    h_node.attrs['np_dtype'] = py_obj.dtype.str.encode('ascii')
    return h_node,(("data",py_obj.data,{},kwargs),('mask',py_obj.mask,{},kwargs))


def load_np_dtype_dataset(h_node,base_type,py_obj_type):
    """
    restores dtype from dataset

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (numpy.dtype):
        final type of restored dtype

    Returns
    -------
    resulting numpy.dtype
    """
    return np.dtype(bytes(h_node[()]))


def load_np_scalar_dataset(h_node,base_type,py_obj_type):
    """
    restores scalar value from dataset

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (numpy.dtype):
        final type of restored dtype

    Returns
    -------
    resulting numpy.scalar
    """

    dtype = np.dtype(h_node.attrs["np_dtype"])
    return dtype.type(h_node[()])


def load_ndarray_dataset(h_node,base_type,py_obj_type):
    """
    restores ndarray like object from dataset

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (numpy.ndarray, numpy.ma.array, ...):
        final type of restored array

    Returns
    -------
    resulting numpy.ndarray, numpy.ma.array
    """
    dtype = np.dtype(h_node.attrs['np_dtype'])
    if "str" in dtype.name:
        string_data = h_node[()]
        if h_node.dtype.itemsize <= 1 or 'bytes' not in h_node.dtype.name:
            # in hickle 4.0.X numpy.ndarrays containing multiple strings are 
            # not converted to list of string but saved as ar consequently
            # itemsize of dtype is > 1
            string_data = bytes(string_data).decode("utf8")
        return np.array(string_data,copy=False,dtype=dtype)
    if issubclass(py_obj_type,np.matrix):
        return py_obj_type(data=h_node[()],dtype=dtype)
    # TODO how to restore other ndarray derived object_types
    # simply using classname for casting does not work, in
    # case they use the same interface like numpy.ndarray
    return np.array(h_node[()], dtype=dtype)


def load_ndarray_masked_dataset(h_node,base_type,py_obj_type):
    """
    restores masked array from data and mask datasets as stored by
    hickle version 4.0.0

    Parameters
    ----------
    h_node (h5py.Dataset):
        the hdf5 node to load data from

    base_type (bytes):
        bytes string denoting base_type

    py_obj_type (numpy.ndarray, numpy.ma.array, ...):
        final type of restored array

    Returns
    -------
    resulting numpy.ndarray, numpy.ma.array
    """
    masked_array = NDMaskedArrayContainer(h_node.attrs,base_type,py_obj_type)
    masked_array.append('data',h_node[()],h_node.attrs),
    mask_path = "{}_mask".format(h_node.name)
    h_root = h_node.parent
    h_node_mask = h_root.get(mask_path,None)
    if h_node_mask is None:
        raise ValueError("mask not found")
    masked_array.append('mask',h_node_mask,h_node_mask.attrs)
    return masked_array.convert()

class NDArrayLikeContainer(ListLikeContainer):
    """
    PyContainer used to restore complex ndarray from h5py.Group node
    """

    __slots__ = ()
    
    def append(self,name,item,h5_attrs):

        # if group contains only one item which either has been
        # dumped using create_pickled_dataset or its name reads
        # data than assume single non list-type object otherwise
        # pass item on to append method of ListLikeContainer
        if h5_attrs.get("base_type",'') == b'pickle' or name == "data":
            self._content = item
        else:
            super(NDArrayLikeContainer,self).append(name,item,h5_attrs)
    
    def convert(self):
        data = np.array(self._content,dtype = self._h5_attrs['np_dtype'])
        return data if data.__class__ is self.object_type or isinstance(self.object_type,types.LambdaType) else self.object_type(data)

class NDMaskedArrayContainer(PyContainer):
    """
    PyContainer used to restore masked array stored as dedicated h5py.Group
    """
    __slots__ = ()

    def __init__(self,h5_attrs,base_type,object_type):
        super(NDMaskedArrayContainer,self).__init__(h5_attrs,base_type,object_type,_content = {})

    def append(self,name,item,h5_attrs):
        self._content[name] = item

    def convert(self):
        dtype = self._h5_attrs['np_dtype']
        data = np.ma.array(self._content['data'], mask=self._content['mask'], dtype=dtype)
        return data if data.__class__ is self.object_type or isinstance(self.object_type,types.LambdaType) else self.object_type(data)

#####################
# Lookup dictionary #
#####################

# %% REGISTERS
class_register = [
    [np.dtype, b"np_dtype", create_np_dtype, load_np_dtype_dataset],
    [np.number, b"np_scalar", create_np_scalar_dataset, load_np_scalar_dataset,None,False],

    # for all scalars which are not derived from numpy.number which itself is numpy.generic subclass
    # to properly catch and handle they will be caught by the following
    [np.generic, b"np_scalar", create_np_scalar_dataset, load_np_scalar_dataset,None,False],

    [np.ndarray, b"ndarray", create_np_array_dataset, load_ndarray_dataset,NDArrayLikeContainer],
    [np.ma.core.MaskedArray, b"ndarray_masked", create_np_masked_array_dataset, None,NDMaskedArrayContainer],

    # NOTE: The following is load only
    #       just needed to link old ndarray_masked_data base_type to load_ndarray_masked_dataset
    #       loader module selection will be triggered by numpy.ma.core.MaskedArray object_type anyway
    #       but base_type is used to select proper load_function
    [np.ma.core.MaskedArray, b"ndarray_masked_data",None , load_ndarray_masked_dataset,None,False,'hickle-4.x'],

    # NOTE: numpy.matrix is obsolete and just an alias for numpy.ndarray therefore 
    # to keep things simple numpy.matrix will be handled by same functions as 
    # numpy.ndarray. As long as just required cause numpy.ma.core.MaskedArray 
    # uses it for data
    [np.matrix, b"np_matrix", create_np_array_dataset, load_ndarray_dataset]
]

exclude_register = [
    (b"ndarray_masked_mask",'hickle-4.x')
]



"""
# Generate_test_hickle.py

Helper to generate test hickle files for a given hickle version

Bash command to generate things:
> VER=2.1.0; pip uninstall hickle -y; pip install hickle==$VER; python generate_test_hickle.py $VER
"""
import hickle as hkl
import numpy as np
import sys

ver_str = sys.argv[1].replace('.', '_')

fn_out = 'hickle_%s.hkl' % ver_str

dd = {
    'dog_breed': b'Chihuahua',
    'age': 10,
    'height': 1.1,
    'nums': [1, 2, 3],
    'narr': np.array([1, 2, 3]),
}

print("Dumping %s..." % fn_out)
hkl.dump(dd, fn_out, path='test')

